package tk.to_p.blankedoutbox.garlicclone.ai;


import java.util.Vector;

import tk.to_p.blankedoutbox.garlicclone.core.GamePlay;
import tk.to_p.blankedoutbox.garlicclone.core.GarlicConst;
import tk.to_p.blankedoutbox.garlicclone.core.Player;
import tk.to_p.blankedoutbox.garlicclone.core.Playfield;
import tk.to_p.blankedoutbox.garlicclone.core.TetriMinos;

/**
 * AI codes
 */
public class TetrisAI {
	/** AI state: Thinking */
	public static final int STATE_THINKING = 1;
	/** AI state: Think start */
	public static final int STATE_THINK_START = 2;
	/** AI state: Moving */
	public static final int STATE_MOVING = 3;
	/** Leftmost X coord */
	public static final int BLOCK_MIN_LEFT = -2;
	/** Index ID for nextBlock_ variable: Hold piece */
	public static final int HOLD_BLOCK = 0;
	/** Index ID for nextBlock_ variable: Current piece */
	public static final int CUR_BLOCK = 1;
	/** Minimum evaluation points */
	public static final int EVAL_MIN = -200000;

	/** Think delay */
	public int reqThinkDelayFrame_ = 0;
	/** Current think delay timer */
	public int curThinkDelayFrame_ = 0;
	/** Move delay */
	public int reqMoveDelayFrame_ = 0;
	/** Current move delay timer */
	public int curMoveDelayFrame_ = 0;
	/** true if hold is available */
	public boolean canHold_ = true;
	/** An array of next pieces */
	public Vector<Integer> nextBlock_ = new Vector<Integer>();
	/** Something related to hold */
	public int holdMinDiff_ = 0;
	/** T-Spin check enable flag... but they didn't port the T-Spin codes correctly */
	public boolean enableTSpin_ = false;
	/** Something related to T-Spins */
	public int tspinMinDiff_ = 0;
	/** Best move data */
	public MoveData bestMove_ = new MoveData();

	/** Bonus for HighestBlock */
	public int coHighestBlock_ = -10;
	/** Bonus for HorizontalTransitions */
	public int coHorizontalTransitions_ = -10;
	/** Bonus for VerticalTransitions */
	public int coVerticalTransitions_ = -10;
	/** Bonus for BlockedCells */
	public int coBlockedCells_ = -10;
	/** Bonus for Wells */
	public int coWells_ = -9;
	/** Bonus for FilledLines */
	public int coFilledLines_ = 1;
	/** Bonus for BlockedWells */
	public int coBlockedWells_ = -10;

	/** GamePlay that this AI runs on */
	public GamePlay gamePlay_;
	/** Playfield Width */
	public int width_;
	/** Playfield Height */
	public int height_;
	/** AI's virtual field */
	public boolean[] fields_;
	/** Current AI status */
	public int curState_;

	/** Current highest block height */
	public int curBlockHighest_;
	/** Current concrete block height */
	public int curConcreteHeight_;
	/** Current playable height */
	public int curBlockLastHeight_;

	/** Currently moving piece ID */
	public int curMovingPieceId_;

	/**
	 * [Added] Default Constructor
	 */
	public TetrisAI() {
	}

	/**
	 * [Added] Constructor
	 * @param gamePlay GamePlay
	 */
	public TetrisAI(GamePlay gamePlay) {
		Initialize(gamePlay);
	}

	/**
	 * Initialize the AI.
	 * @param gamePlay GamePlay
	 * @return The return value is determined by DoInitialize(), which is undefined in flash games.
	 */
	public boolean Initialize(GamePlay gamePlay) {
		System.out.println("TetrisAI::Initialize");
		gamePlay_ = gamePlay;
		width_ = gamePlay_.GetPlayerSetting().PlayfieldWidth;
		height_ = gamePlay_.GetPlayerSetting().PlayfieldHeight;
		fields_ = new boolean[width_ * height_];
		curState_ = STATE_THINK_START;
		return DoInitialize();
	}

	/**
	 * Shutdown the AI.
	 * The only thing it does is execute a function named DoShutdown(), but it's not defined in flash games.
	 */
	public void Shutdown() {
		DoShutdown();
	}

	/**
	 * Test the Tetrimino collision in the AI's virtual-field.
	 * @param x X position
	 * @param y Y position
	 * @param tetriMino Tetrimino to test the collision
	 * @return Returns <code>true</code> if the Tetrimino does <b>NOT</b> overlaps to the existing blocks.
	 *         In other words, it returns <code>true</code> if the Tetrimino can be moved to the specified position.
	 */
	public boolean TestShape(int x, int y, TetriMinos.TetriMino tetriMino) {
		for(int i = 0; i < GarlicConst.Tetri; i++) {
			int x2 = x + tetriMino.minos_[i].x_;
			int y2 = y + tetriMino.minos_[i].y_;
			if (!(x2 < width_ && y2 < height_ && x2 >= 0 && y2 >= 0)) {
				return false;
			} else if (fields_[y2 * width_ + x2]) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Add/remove the Tetrimino-shape to the AI's virtual-field.
	 * If used to empty spaces, it will place the Tetrimino. If used to non-empty space, it will create the Tetrimino-shaped hole.
	 * @param x X position
	 * @param y Y position
	 * @param tetriMino Tetrimino to add/remove
	 */
	public void ToggleShape(int x, int y, TetriMinos.TetriMino tetriMino) {
		for(int i = 0; i < GarlicConst.Tetri; i++) {
			int x2 = x + tetriMino.minos_[i].x_;
			int y2 = y + tetriMino.minos_[i].y_;
			fields_[y2 * width_ + x2] = !fields_[y2 * width_ + x2];
		}
	}

	/**
	 * Prepare the think
	 */
	public void PrepareThink() {
		int MAX_NEXT = 5;
		nextBlock_.clear();

		for(int i = 0; i < MAX_NEXT; i++) {
			int pieceType = GarlicConst.TetriminoType_None;

			if(i == 0) { // HOLD_BLOCK
				pieceType = gamePlay_.GetPlayer().tetriMinoInHold_;
				if(pieceType == GarlicConst.TetriminoType_None) {
					pieceType = gamePlay_.GetPlayer().GetTetriminoInBag(0);
				}
			} else if(i == 1) { // CUR_BLOCK
				pieceType = gamePlay_.GetPlayer().GetCurPiece().type_;
			} else {
				pieceType = gamePlay_.GetPlayer().GetTetriminoInBag(i - 2);
			}

			if(pieceType != GarlicConst.TetriminoType_None) {
				nextBlock_.add(pieceType);
			}
		}

		int cy = gamePlay_.GetPlayfield().GetConcreteLineCnt();
		curBlockHighest_ = 0;

		for(int i = 0; i < gamePlay_.GetPlayfield().GetWidth(); i++) {
			int y = gamePlay_.GetPlayfield().height_ - 1 - cy;
			while(y >= 0) {
				Playfield.Grid grid = gamePlay_.GetPlayfield().GetGrid(i, y);
				boolean isExist = !grid.empty_;

				if(isExist) {
					int y2 = gamePlay_.GetPlayfield().height_ - y;
					curBlockHighest_ = Math.max(y2, curBlockHighest_);
				}

				fields_[i + (y + cy) * width_] = isExist;
				y--;
			}
		}

		curConcreteHeight_ = cy;
		curBlockLastHeight_ = gamePlay_.GetPlayfield().GetPlayableHeight();
		canHold_ = gamePlay_.GetPlayer().CanHold();
	}

	/**
	 * Most top-level function of think routines, but the actual logic is elsewhere.
	 * @return <code>true</code> if successful
	 */
	public boolean Think() {
		boolean thinkSuccess = DoThink(bestMove_);
		//int cy = gamePlay_.GetPlayfield().GetConcreteLineCnt(); // Does seem unused
		bestMove_.y0 -= curConcreteHeight_;
		bestMove_.y0 -= curConcreteHeight_;
		return thinkSuccess;
	}

	/**
	 * The main routine of the AI. Should be called before the Garlic engine's game play updates.
	 */
	public void Run() {
		// What's this? GetCurPlayInfo() is undefined in the Flash games.
		//GetCurPlayInfo().curComboCnt_ = gamePlay_.GetStatistics().curComboCnt_;

		if(curState_ == STATE_THINK_START)
		{
			Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

			if(curPiece != null) {
				PrepareThink();
				curMovingPieceId_ = curPiece.id_;
				curState_ = STATE_THINKING;
				curThinkDelayFrame_ = reqThinkDelayFrame_;
			}
		}
		if(curState_ == STATE_THINKING)
		{
			if(--curThinkDelayFrame_ <= 0) {
				gamePlay_.SetControllerFlags(0);
				Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

				if(curPiece != null) {
					if (curMovingPieceId_ != curPiece.id_) {
						curState_ = STATE_THINK_START;
					} else if(Think()) {
						curState_ = STATE_MOVING;
						curMoveDelayFrame_ = reqMoveDelayFrame_;
					}
				}
				return;
			}
		}
		if(curState_ == STATE_MOVING)
		{
			int HOLDACTION_NONE = 0;
			int HOLDACTION_EXECUTE = 1;
			//int HOLDACTION_UNUSED = 2;

			if(curMoveDelayFrame_-- > 0) {
				gamePlay_.SetControllerFlags(0);
			} else {
				curMoveDelayFrame_ = reqMoveDelayFrame_;
				int ctrlFlag = 0;
				Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();

				if(curPiece != null)
				{
					if(bestMove_.hold == HOLDACTION_EXECUTE) {
						if (curMovingPieceId_ != curPiece.id_) {
							bestMove_.hold = HOLDACTION_NONE;
							curMovingPieceId_ = curPiece.id_;
						} else {
							ctrlFlag = GarlicConst.ControllerFlag_HOLD;
						}
					} else {
						if(curMovingPieceId_ != curPiece.id_) {
							curState_ = STATE_THINK_START;
						} else {
							int curX;
							int curY;
							curX = curPiece.x_;
							curY = curPiece.y_;

							int xn;
							int an;
							if(curY < bestMove_.y0) {
								xn = bestMove_.x0;
								an = bestMove_.a0;
							} else {
								xn = bestMove_.x1;
								an = bestMove_.a1;
							}

							if (curX < xn) {
								ctrlFlag |= GarlicConst.ControllerFlag_RIGHT;
							} else if (curX > xn) {
								ctrlFlag |= GarlicConst.ControllerFlag_LEFT;
							}

							if (an != curPiece.face_) {
								int face = an - curPiece.face_;
								if(face < 0) face += GarlicConst.FaceType_Count;

								if(face > GarlicConst.FaceType_Count / 2) {
									ctrlFlag |= GarlicConst.ControllerFlag_CCW;
								} else {
									ctrlFlag |= GarlicConst.ControllerFlag_CW;
								}
							}

							if (bestMove_.a0 == bestMove_.a1
							 && bestMove_.x0 == bestMove_.x1
							 && curX == xn && an == curPiece.face_)
							{
								ctrlFlag |= GarlicConst.ControllerFlag_HARD;
								curState_ = STATE_THINK_START;
							}
							else if (curX == xn && an == curPiece.face_)
							{
								ctrlFlag |= GarlicConst.ControllerFlag_SOFT;
							}
						}
					}
				} /* if(curPiece != null) */

				gamePlay_.SetControllerFlags(ctrlFlag);
				if(curState_ == STATE_MOVING) {
					curMoveDelayFrame_ = Math.max(1, curMoveDelayFrame_);
				}
			}
		} /* if(curState_ == STATE_MOVING) */
	}

	/**
	 * Modify the status of AI's virtual field. It doesn't check the bounds.
	 * @param x X position
	 * @param y Y position
	 * @param exist <code>true</code> if a block exist
	 * @throws ArrayIndexOutOfBoundsException If the specified position is invalid
	 */
	public void SetField(int x, int y, boolean exist) {
		fields_[y * width_ + x] = exist;
	}

	/**
	 * Get the status of AI's virtual field. If out-of-bounds position is specified, it will return <code>true</code>.
	 * @param x X position
	 * @param y Y position
	 * @return <code>true</code> if a block exist, or if you specify invalid position
	 */
	public boolean GetField(int x, int y) {
		if (x < 0 || y < 0 || x >= width_ || y >= height_) {
			return true;
		}
		return fields_[y * width_ + x];
	}

	/**
	 * Create a mysterious "pattern number" based on current state of the virtual field. Used for T-Spin detections.
	 * @param x X position
	 * @param y Y position
	 * @param width Pattern Width
	 * @param height Pattern Height
	 * @return A mysterious "pattern number". If (width * height) is bigger than 32, it will return 0.
	 */
	public int GetMatrixPattern(int x, int y, int width, int height) {
		if(width * height > 32) return 0;

		int ptn = 0;
		int cnt = width * height - 1;

		for(int i = y; i < y + height; i++) {
			for(int j = x; j < x + width; j++) {
				ptn |= ((GetField(j, i) ? 1 : 0) << cnt);
				cnt--;
			}
		}

		return ptn;
	}

	/**
	 * Evaluate the next piece.
	 * @param curDepth Depth
	 * @param result The results will be stored to this object
	 */
	public void ThinkNextBlock(int curDepth, ThinkNextBlockResult result) {
		result.best = EVAL_MIN;

		for(int i = 0; i < 4; i++) {
			TetriMinos.TetriMino tetriMino = GarlicConst.TetriMinos_[nextBlock_.get(curDepth)].tetriMinos_[i];

			for(int x = BLOCK_MIN_LEFT; x < width_; x++) {
				int y = height_ / 2;
				for(;;) {
					if(!(TestShape(x, y, tetriMino) && y < height_)) break;
					y++;
				}
				y--;

				if(TestShape(x, y, tetriMino)) {
					ToggleShape(x, y, tetriMino);
					int points = Evalfield(curDepth);
					ToggleShape(x, y, tetriMino);

					if(result.best < points) {
						result.best = points;
						result.aBest = i;
						result.xBest = x;
						result.yBest = y;
					}
				}
			}
		}
	}

	/**
	 * Think about various things
	 * @param moveData MoveData (The results will be stored here)
	 * @return It seems it always returns true...
	 */
	public boolean DoThink(MoveData moveData) {
		ThinkNextBlockResult thinkResult = new ThinkNextBlockResult();
		boolean curPieceIsT = (nextBlock_.get(CUR_BLOCK) == GarlicConst.TetriminoType_T);
		boolean nextPieceIsT = (nextBlock_.get(CUR_BLOCK + 1) == GarlicConst.TetriminoType_T);
		boolean holdIsT = canHold_ && (nextBlock_.get(HOLD_BLOCK) == GarlicConst.TetriminoType_T);
		boolean isHold = false;
		ThinkNextBlockResult thinkResultNext = new ThinkNextBlockResult();
		ThinkNextBlockResult thinkResultHold = new ThinkNextBlockResult();

		ThinkNextBlock(CUR_BLOCK, thinkResultNext);
		if(canHold_) ThinkNextBlock(HOLD_BLOCK, thinkResultHold);

		if(thinkResultHold.best > thinkResultNext.best + holdMinDiff_) {
			thinkResult = thinkResultHold;
			isHold = true;
		} else {
			thinkResult = thinkResultNext;
		}

		if(enableTSpin_) {
			if(curPieceIsT || holdIsT) {
				MoveData tSpinPosResult = new MoveData();
				boolean isTSpinPosFound = FindTSpinPos(BLOCK_MIN_LEFT, width_, tSpinPosResult);

				if(holdIsT && isTSpinPosFound) {
					MakeTSpinMatrixRetObject tMatrixResult = new MakeTSpinMatrixRetObject();
					if(MakeTSpinMatrix(CUR_BLOCK, true, tMatrixResult) && tMatrixResult.best > tSpinPosResult.best) {
						moveData.x1 = tMatrixResult.x;
						moveData.x0 = moveData.x1;
						moveData.y1 = tMatrixResult.y;
						moveData.y0 = moveData.y1;
						moveData.a1 = tMatrixResult.a;
						moveData.a0 = moveData.a1;
						moveData.hold = 0;
						return true;
					}
				}

				if(canHold_ && curPieceIsT && isTSpinPosFound) {
					MakeTSpinMatrixRetObject tMatrixResult = new MakeTSpinMatrixRetObject();
					if(MakeTSpinMatrix(HOLD_BLOCK, true, tMatrixResult) && tMatrixResult.best > tSpinPosResult.best) {
						moveData.x1 = tMatrixResult.x;
						moveData.x0 = moveData.x1;
						moveData.y1 = tMatrixResult.y;
						moveData.y0 = moveData.y1;
						moveData.a1 = tMatrixResult.a;
						moveData.a0 = moveData.a1;
						moveData.hold = 1;
						return true;
					}
				}

				// XXX:best and info are undefined in Flash games
				int best = 0;
				MoveData info = new MoveData();
				if (isTSpinPosFound && tspinMinDiff_ + best > thinkResult.best) {
					moveData.x0 = info.x0;
					moveData.x1 = info.x1;
					moveData.y0 = info.y0;
					moveData.y1 = info.y1;
					moveData.a0 = info.a0;
					moveData.a1 = info.a1;
					moveData.hold = (!curPieceIsT && holdIsT) ? 1 : 0;
					return true;
				}
			}

			if(nextPieceIsT || holdIsT) {
				MakeTSpinMatrixRetObject info = new MakeTSpinMatrixRetObject();
				int best = 0;	// XXX:undefined again
				if(MakeTSpinMatrix(CUR_BLOCK, false, info) && best > thinkResult.best) {
					moveData.x1 = info.x;
					moveData.x0 = moveData.x1;
					moveData.y1 = info.y;
					moveData.y0 = moveData.y1;
					moveData.a1 = info.a;
					moveData.a0 = moveData.a1;
					moveData.hold = 0;
					return true;
				}
			}

			if(canHold_ && !isHold && curPieceIsT && !holdIsT && thinkResultHold.best + holdMinDiff_ > thinkResultNext.best) {
				thinkResult = thinkResultHold;
				isHold = true;
			}
		}

		moveData.x0 = thinkResult.xBest;
		moveData.x1 = thinkResult.xBest;
		moveData.y0 = thinkResult.yBest;
		moveData.y1 = thinkResult.yBest;
		moveData.a0 = thinkResult.aBest;
		moveData.a1 = thinkResult.aBest;
		moveData.hold = (isHold ? 1 : 0);

		return true;
		//return false;
	}

	/**
	 * Get the number of lines filled by putting a piece
	 * @param m The TetriMino to put
	 * @param x X position
	 * @param y Y position
	 * @return Number of lines
	 */
	public int GetFilledLine(TetriMinos.TetriMino m, int x, int y) {
		for(;;) {
			if (!(TestShape(x, y, m) && y < height_)) break;
			y++;
		}
		y--;

		if(TestShape(x, y, m)) {
			ToggleShape(x, y, m);
			int lines = 0;

			for(int i = height_ / 2; i < height_; i++) {
				int j = 0;
				while(j < width_) {
					if(!fields_[i * width_ + j]) {
						break;
					}
					j++;
				}
				if(j == width_) {
					lines++;
				}
			}

			ToggleShape(x, y, m);
			return lines;
		}

		return 0;
	}

	/**
	 * This function will use the mysterious "pattern number" (created by GetMatrixPattern) to check for possible T-Spins.
	 * @param pattern The mysterious "pattern number"
	 * @param objReturn The results will be stored here
	 * @return <code>true</code> if a T-Spin is found
	 */
	public boolean IsTSpinPattern(int pattern, IsTSpinPatternRetObject objReturn) {
		int FACE_NORTH = 0;
		int FACE_EAST = 1;
		int FACE_SOUTH = 2;
		int FACE_WEST = 3;

		switch (pattern) {
		case 8219:
		case 8251:
		case 9243:
		case 9275:
		case 24603:
		case 24635:
		case 25627:
		case 25659:
		case 8731:
		case 8763:
		case 9755:
		case 9787:
		case 25115:
		case 25147:
		case 26139:
		case 26171:
			objReturn.a0 = FACE_EAST;
			objReturn.a1 = FACE_SOUTH;
			objReturn.xDiff = 0;
			objReturn.yDiff = 0;
			return true;
		case 2075:
		case 2107:
		case 2587:
		case 2619:
		case 3099:
		case 3131:
		case 3611:
		case 3643:
		case 18459:
		case 18491:
		case 18971:
		case 19003:
		case 19483:
		case 19515:
		case 19995:
		case 20027:
			objReturn.a0 = FACE_WEST;
			objReturn.a1 = FACE_SOUTH;
			objReturn.xDiff = 0;
			objReturn.yDiff = 0;
			return true;
		case 1085:
		case 17469:
		case 17981:
		case 1597:
			objReturn.a0 = FACE_NORTH;
			objReturn.a1 = FACE_WEST;
			objReturn.xDiff = 1;
			objReturn.yDiff = 0;
			return true;
		case 16919:
		case 17975:
		case 17943:
		case 16951:
			objReturn.a0 = FACE_NORTH;
			objReturn.a1 = FACE_EAST;
			objReturn.xDiff = -1;
			objReturn.yDiff = 0;
			return true;
		case 312:
		case 824:
		case 1336:
		case 1848:
		case 8504:
		case 9016:
		case 9528:
		case 10040:
		case 16696:
		case 17208:
		case 17720:
		case 18232:
		case 24888:
		case 25400:
		case 25912:
		case 26424:
			objReturn.a0 = FACE_EAST;
			objReturn.a1 = FACE_NORTH;
			objReturn.xDiff = 1;
			objReturn.yDiff = 1;
			return true;
		case 579:
		case 611:
		case 1603:
		case 1635:
		case 2627:
		case 2659:
		case 3651:
		case 3683:
		case 16963:
		case 16995:
		case 17987:
		case 18019:
		case 19011:
		case 19043:
		case 20035:
		case 20067:
			objReturn.a0 = FACE_WEST;
			objReturn.a1 = FACE_NORTH;
			objReturn.xDiff = -1;
			objReturn.yDiff = 1;
			return true;
		}

		return false;
	}

	/**
	 * This function will try to find the possible T-Spins... but they didn't port this well.
	 * @param xMin Minimum X position
	 * @param xMax Maximum X position
	 * @param objReturn The results will be stored to this thing
	 * @return <code>true</code> if a T-Spin is found
	 */
	public boolean FindTSpinPos(int xMin, int xMax, MoveData objReturn) {
		objReturn.best = EVAL_MIN;
		boolean result = false;

		for(int face = 0; face < 4; face++) {
			if(face == 2) {
				// Do Nothing
			} else {
				TetriMinos.TetriMino tetriMino = GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[face];

				for(int x = xMin; x < xMax; x++) {
					int y = height_ / 2;
					for(;;) {
						if(!(TestShape(x, y, tetriMino) && y < height_)) break;
						y++;
					}
					y--;

					if(TestShape(x, y, tetriMino)) {
						int pattern = GetMatrixPattern(x - 1, y, 5, 3);
						int a0 = 0; // XXX:no default value defined
						int a1 = 0; // XXX:no default value defined
						//int unused1, unused2;
						IsTSpinPatternRetObject tPatternResult = new IsTSpinPatternRetObject();

						if(IsTSpinPattern(pattern, tPatternResult) && (tPatternResult.a0 == face)) {
							int x2 = x + tPatternResult.xDiff;
							int y2 = y + tPatternResult.yDiff;
							ToggleShape(x2, y2, GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[tPatternResult.a1]);

							int cornerCount = 0;
							if(GetField(x2    , y2    )) cornerCount++;
							if(GetField(x2 + 2, y2    )) cornerCount++;
							if(GetField(x2    , y2 + 2)) cornerCount++;
							if(GetField(x2 + 2, y2 + 2)) cornerCount++;

							boolean isProperSpin = false;
							if (GarlicConst.FaceType_North == face && GetField(x2 + 1, y2 + 2) ||
								GarlicConst.FaceType_East  == face && GetField(x2    , y2 + 1) ||
								GarlicConst.FaceType_South == face && GetField(x2 + 1, y2    ) ||
								GarlicConst.FaceType_West  == face && GetField(x2 + 2, y2 + 1))
							{
								isProperSpin = true;
							}

							if(cornerCount >= 3) {
								int noHoleCount = 0;

								for(int y3 = 0; y3 < 3; y3++) {
									if(y2 + y3 >= height_) break;

									boolean isNoHoles = true;
									for(int x3 = 0; x3 < width_; x3++) {
										if(!GetField(x3, y2 + y3)) {
											isNoHoles = false;
											break;
										}
									}

									if(isNoHoles) noHoleCount++;
								}

								if(noHoleCount > 0) {
									int tspinMinDiff = tspinMinDiff_;
									int points = Evalfield(0) + (noHoleCount - (isProperSpin ? 1 : 0)) * tspinMinDiff;
									int best = 0;	// XXX:best is not defined in Flash games. They obviously didn't fully port the T-Spin codes...
									if(points > best) {
										objReturn.best = points;
										objReturn.x0 = x;
										objReturn.x1 = x;
										objReturn.y0 = y;
										objReturn.y1 = y;
										objReturn.a0 = a0;
										objReturn.a1 = a1;
										result = true;
									}
								}
							}

							// XXX:This "a1" is defined in above, but has undefined value, so it doesn't work well
							ToggleShape(x2, y2, GarlicConst.TetriMinos_[GarlicConst.TetriminoType_T].tetriMinos_[a1]);
						}
					}
				}
			}
		}

		return result;
	}

	/**
	 * Something to do with the T-Spins
	 * @param curDepth Depth
	 * @param checkAll <code>true</code> to check all X positions
	 * @param objReturn The results will be stored here
	 * @return <code>true</code> if a T-Spin is found
	 */
	public boolean MakeTSpinMatrix(int curDepth, boolean checkAll, MakeTSpinMatrixRetObject objReturn) {
		objReturn.best = EVAL_MIN;
		boolean result = false;

		for(int i = 0; i < 4; i++) {
			TetriMinos.TetriMino tetriMino = GarlicConst.TetriMinos_[nextBlock_.get(curDepth)].tetriMinos_[i];

			for(int x = BLOCK_MIN_LEFT; x < width_; x++) {
				int y = height_ / 2;
				for(;;) {
					if(!(TestShape(x, y, tetriMino) && y < height_)) break;
					y++;
				}
				y--;

				if(TestShape(x, y, tetriMino)) {
					ToggleShape(x, y, tetriMino);
					MoveData tSpinPosResult = new MoveData();
					boolean tSpinPosSuccess = false;
					if(checkAll) {
						tSpinPosSuccess = FindTSpinPos(BLOCK_MIN_LEFT, width_, tSpinPosResult);
					} else {
						// Those tetriMino.xMin_ and tetriMino.xMax_ didn't have a underbar at the end
						tSpinPosSuccess = FindTSpinPos(x + tetriMino.xMin_ + BLOCK_MIN_LEFT, x + tetriMino.xMax_, tSpinPosResult);
					}
					ToggleShape(x, y, tetriMino);

					if(tSpinPosSuccess) {
						int e = 0;		// XXX:e is not defined in Flash games
						int best = 0;	// XXX:and this one too...
						if (e > best) {
							// XXX:So this block is a dead code
							objReturn.best = e;
							objReturn.a = i;
							objReturn.x = x;
							objReturn.y = y;
							result = true;
						}
					}
				}
			}
		}

		return result;
	}

	/**
	 * Evaluate the entire field
	 * @param depth Depth
	 * @return Score of this entire field
	 */
	public int Evalfield(int depth) {
		return EvalfieldRange(depth, 0, width_);
	}

	/**
	 * Evaluate the field. This is where the actual brains of AI is stored.
	 * @param depth Depth
	 * @param left Minimum X position
	 * @param right Maximum X position
	 * @return Score of this field
	 */
	public int EvalfieldRange(int depth, int left, int right) {
		// Intermediate Results
		int HorizontalTransitions = 0;
		int FilledLines = 0;
		int HighestBlock = -1;
		int VerticalTransitions = 0;
		int BlockedCells = 0;
		int Wells = 0;
		int StartScore = 0;
		int BlockedWells = 0;

		int x;
		int y;
		int i;
		boolean full;
		boolean empty;
		boolean btype;
		//int well;
		boolean blocked;
		boolean lblock;
		boolean rblock;
		int width;
		int height;
		//int startHeight;

		width = width_;
		height = curBlockLastHeight_;
		boolean[] efield = new boolean[width_ * height_];
		HighestBlock = height;

		// Copy to eField, Calculate HorizontalTransitions, FilledLines and HighestBlock
		y = height - 1;
		i = y;
		while (y >= 0) {
			btype = true; full = true; empty = true;
			x = width - 1;

			while (x >= 0) {
				if (!fields_[y * width + x]) {
					efield[x + i * width] = false;
				} else {
					efield[x + i * width] = true;
					empty = false;
				}

				if(btype != efield[x + i * width]) {
					btype = efield[x + i * width];
					full = false;
					HorizontalTransitions++;
				}

				x--;
			}

			if (!full) {
				i--;
			} else {
				FilledLines++;
			}

			if (!btype) {
				HorizontalTransitions++;
			}

			if (empty) {
				HighestBlock = height - (i + 2);
				break;
			}

			y--;
		}

		x = left;
		while (x < right) {
			btype = false; blocked = false;
			int blockedHeight = 0;

			y = height - HighestBlock;
			while (y < height) {
				if (btype != efield[x + y * width]) {
					if (!blocked) {
						blocked = true;
						blockedHeight = y;
					}
					btype = efield[x + y * width];
					VerticalTransitions++;
				}

				if (blocked && !efield[x + y * width]) {
					BlockedCells++;
					BlockedWells += y - blockedHeight;
				}

				if (x == 0) {
					lblock = true;
				} else {
					lblock = efield[x - 1 + y * width];
				}

				if (x == width - 1) {
					rblock = true;
				} else {
					rblock = efield[x + 1 + y * width];
				}

				if (lblock && rblock) {
					i = y;
					while (i < height) {
						if (efield[x + i * width]) {
							break;
						}
						Wells++;
						i++;
					}
				}
				y++;
			}
			if (!btype) {
				VerticalTransitions++;
			}
			x++;
		}

		int FinalScore = 0;
		FinalScore =
			StartScore
			+ coHighestBlock_ * HighestBlock
			+ coHorizontalTransitions_ * HorizontalTransitions
			+ coVerticalTransitions_ * VerticalTransitions
			+ coBlockedCells_ * BlockedCells
			+ coWells_ * Wells
			+ coFilledLines_ * FilledLines
			+ coBlockedWells_ * BlockedWells;
		return FinalScore;
	}

	/**
	 * Evaluate the piece placement (used by MissDrop checker)
	 * @param m TetriMino
	 * @param x X position
	 * @param y Y position
	 * @return Score of this placement
	 */
	public int EvalBlock(TetriMinos.TetriMino m, int x, int y) {
		int points1 = EvalfieldRange(0, x + m.xMin_, x + m.xMax_);

		for (;;) {
			if (!(TestShape(x, y, m) && y < height_)) break;
			y++;
		}
		y--;

		if (TestShape(x, y, m)) {
			ToggleShape(x, y, m);
			int points2 = EvalfieldRange(0, x + m.xMin_, x + m.xMax_);
			ToggleShape(x, y, m);
			return points2 - points1;
		}

		return EVAL_MIN;
	}

	/**
	 * MissDrop checker
	 * @param range Some range?
	 * @param minDiff Something to do with the placement points
	 * @param maxEval If the first calculated score exceeds this, then it's not a miss drop.
	 * @return <code>true</code> if this is a miss drop
	 */
	public boolean IsMissDrop(int range, int minDiff, int maxEval) {
		Player.Piece curPiece = gamePlay_.GetPlayer().GetCurPiece();
		if(curPiece == null) {
			return false;
		}
		PrepareThink();

		int x;
		int y;
		x = curPiece.x_;
		y = curPiece.y_;
		TetriMinos.TetriMino tetriMino = curPiece.GetTetriMino();

		coHighestBlock_ = 0;
		coHighestBlock_ = 0;
		coHorizontalTransitions_ = 0;
		coVerticalTransitions_ = -100;
		coBlockedCells_ = 0;
		coWells_ = 0;
		coFilledLines_ = 100000;
		coBlockedWells_ = 0;

		int points = EvalBlock(tetriMino, x, y);
		if(points > maxEval) {
			return false;
		}

		coFilledLines_ = 0;

		int points2 = EVAL_MIN;
		int x2 = x - range;

		while (x2 <= x + range) {
			if (x2 == x) {
				// Do nothing
			} else {
				int tempPoints = EvalBlock(tetriMino, x2, y);
				points2 = Math.max(tempPoints, points2);
			}
			x2++;
		}

		return (points2 > points + minDiff);
	}

	/**
	 * [Called by other functions, but not defined in flash games]<br>
	 * This is called at the end of Initialize routine.<br>
	 * I think the "Start" function of TOJ's Lua scripts is called from here.
	 * @return SOMETHING... but I think it's a boolean value. Because this function is just a one of the initialize routines.
	 */
	public boolean DoInitialize() {
		return true;
	}

	/**
	 * [Called by other functions, but not defined in flash games]<br>
	 * This is called from Shutdown routine.<br>
	 * I think the "End" function of TOJ's Lua scripts is called from here.
	 */
	public void DoShutdown() {
	}

	/**
	 * Move data for best movements. Used by DoThink and FindTSpinPos.<br>
	 * [Anonymous class]
	 */
	public static class MoveData {
		public int x0;
		public int x1;
		public int y0;
		public int y1;
		public int a0;
		public int a1;
		public int hold;
		public int best;
	}

	/**
	 * The results of ThinkNextBlock<br>
	 * [Anonymous class]
	 */
	public static class ThinkNextBlockResult {
		public int best;
		public int aBest;
		public int xBest;
		public int yBest;
	}

	/**
	 * The results of IsTSpinPattern<br>
	 * [Anonymous class]
	 */
	public static class IsTSpinPatternRetObject {
		public int a0;
		public int a1;
		public int xDiff;
		public int yDiff;
	}

	/**
	 * The results of MakeTSpinMatrix<br>
	 * [Anonymous class]
	 */
	public static class MakeTSpinMatrixRetObject {
		public int best;
		public int a;
		public int x;
		public int y;
	}
}
