package tk.to_p.blankedoutbox.garlicclone.core;

/**
 * Button input
 */
public class Controller {
	/** GamePlay that owns this Controller */
	public GamePlay gamePlay_ = null;

	/** Button status array */
	public Buttons[] btns_ = new Buttons[GarlicConst.BTN_COUNT];

	/** Button input bit flags */
	public int controllerFlags_ = 0;

	/**
	 * Constructor
	 * @param gamePlay GamePlay that owns this Controller
	 */
	public Controller(GamePlay gamePlay) {
		gamePlay_ = gamePlay;
	}

	/**
	 * Is the button activated?
	 * @param idx Button ID
	 * @return <code>true</code> if the button is activated and some actions should happen
	 */
	public boolean IsBtnActivated(int idx) {
		return btns_[idx].activated_;
	}

	/**
	 * Get the button activated count
	 * @param idx Button ID
	 * @return Button activated count (0=Inactive)
	 */
	public int GetBtnActivatedCnt(int idx) {
		return btns_[idx].activatedCnt_;
	}

	/**
	 * Is the button currently pushed?
	 * @param idx Button ID
	 * @return <code>true</code> if the button is currently pushed
	 */
	public boolean IsBtnPushed(int idx) {
		return btns_[idx].pushed_;
	}

	/**
	 * Initialize
	 */
	public void Init() {
		controllerFlags_ = 0;

		for(int i = 0; i < GarlicConst.BTN_COUNT; i++) {
			btns_[i] = new Buttons();
			btns_[i].pushed_ = false;
			btns_[i].activated_ = false;
			btns_[i].activatedCnt_ = 0;
			btns_[i].remainMsec_ = 0;
			btns_[i].pushedTime_ = 0;
		}
	}

	/**
	 * Set the button input bit flags
	 * @param flags Button input bit flags
	 */
	public void SetControllerFlags(int flags) {
		controllerFlags_ = flags;
	}

	/**
	 * Get action delay or interval
	 * @param btnIdx Button ID
	 * @param first <code>true</code> to get initial delay (such as DAS), <code>false</code> to get further delay (such as ARR)
	 * @return Action delay or interval
	 */
	public int GetButtonRemainDelay(int btnIdx, boolean first) {
		switch(btnIdx) {
		case GarlicConst.BTN_LEFT:
		case GarlicConst.BTN_RIGHT:
			return first ? gamePlay_.GetPlayerSetting().MoveDelay : gamePlay_.GetPlayerSetting().MoveInterval;
		case GarlicConst.BTN_CW:
		case GarlicConst.BTN_CCW:
			return first ? gamePlay_.GetPlayerSetting().RotateDelay : gamePlay_.GetPlayerSetting().RotateInterval;
		case GarlicConst.BTN_SOFT:
			return first ? gamePlay_.GetPlayerSetting().SoftDropDelay : gamePlay_.GetPlayerSetting().SoftDropInterval;
		case GarlicConst.BTN_HARD:
			return first ? gamePlay_.GetPlayerSetting().HardDropDelay : gamePlay_.GetPlayerSetting().HardDropInterval;
		case GarlicConst.BTN_HOLD:
			return 0;
		}
		return 0;
	}

	/**
	 * Main procedure for controller input
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void Run(int msec) {
		for(int i = 0; i < GarlicConst.BTN_COUNT; i++) {
			int bit = 1 << i;

			if((controllerFlags_ & bit) != 0) {
				if(!btns_[i].pushed_) {
					btns_[i].pushed_ = true;
					btns_[i].remainMsec_ = GetButtonRemainDelay(i, true);
					btns_[i].activated_ = true;
					btns_[i].activatedCnt_ = 0;
					btns_[i].pushedTime_ = gamePlay_.GetPlayTime();
				} else {
					btns_[i].activated_ = false;
					if(btns_[i].remainMsec_ > 0) {
						btns_[i].remainMsec_ -= Math.min(msec, btns_[i].remainMsec_);
						if(btns_[i].remainMsec_ <= 0) {
							btns_[i].remainMsec_ = GetButtonRemainDelay(i, false);
							btns_[i].activated_ = true;
							btns_[i].activatedCnt_++;
						}
					}
				}
			} else {
				btns_[i].pushed_ = false;
				btns_[i].activated_ = false;
			}
		}

		if(btns_[GarlicConst.BTN_LEFT].pushed_ && btns_[GarlicConst.BTN_RIGHT].pushed_) {
			if(btns_[GarlicConst.BTN_LEFT].pushedTime_ < btns_[GarlicConst.BTN_RIGHT].pushedTime_) {
				btns_[GarlicConst.BTN_LEFT].activated_ = false;
			} else if(btns_[GarlicConst.BTN_LEFT].pushedTime_ > btns_[GarlicConst.BTN_RIGHT].pushedTime_) {
				btns_[GarlicConst.BTN_RIGHT].activated_ = false;
			} else {
				btns_[GarlicConst.BTN_LEFT].activated_ = false;
				btns_[GarlicConst.BTN_RIGHT].activated_ = false;
			}
		}
	}

	/**
	 * Button input status<br>
	 * [Anonymous class]
	 */
	public static class Buttons {
		/** <code>true</code> if this button is currently pushed */
		public boolean pushed_ = false;

		/** <code>true</code> if this button is activated and some actions should happen */
		public boolean activated_ = false;

		/** Action activated count (0=Inactive) */
		public int activatedCnt_ = 0;

		/** Remaining time before the auto-repeat happens */
		public int remainMsec_ = 0;

		/** Last game-time when the button is pushed */
		public int pushedTime_ = 0;
	}
}
