package tk.to_p.blankedoutbox.garlicclone.core;
import java.util.Random;
import java.util.Vector;

/**
 * This class handles some of the main gameplay.
 */
public class GamePlay {
	/** Playfield */
	public Playfield playfield_ = new Playfield(this);

	/** Settings */
	public PlayerSetting playerSetting_ = new PlayerSetting();

	/** List of EventReceiver */
	public Vector<EventReceiver> eventReceivers_ = new Vector<EventReceiver>();

	/** Score informations */
	public Statistics statistics_ = new Statistics();

	/** Button input */
	public Controller controller_ = new Controller(this);

	/** Player movements */
	public Player player_ = new Player(this);

	/** Incoming garbage lines */
	public Vector<AttackEntry> attackEntries_ = new Vector<AttackEntry>();

	/** Remaining game loop time */
	public int remainLoopTime_ = 0;

	/** <code>true</code> if the game is over */
	public boolean gameOver_ = false;

	/** Ready->Go timer */
	public int curReadyTime_ = 0;

	/** Duration of the Ready->Go status */
	public int reqReadyTime_ = 0;

	/** Elapsed game time */
	public int playTime_ = 0;

	/** <code>true</code> if the game is started */
	public boolean startGame_ = false;

	/** <code>true</code> if the game is paused */
	public boolean pauseGame_ = false;

	/**
	 * Get an instance of Statistics which contains various score informations.
	 * @return An instance of Statistics
	 */
	public Statistics GetStatistics() {
		return statistics_;
	}

	/**
	 * Get an instance of Playfield.
	 * @return An instance of Playfield
	 */
	public Playfield GetPlayfield() {
		return playfield_;
	}

	/**
	 * Get an instance of PlayerSetting.
	 * @return An instance of PlayerSetting
	 */
	public PlayerSetting GetPlayerSetting() {
		return playerSetting_;
	}

	/**
	 * Get an instance of Player.
	 * @return An instance of Player
	 */
	public Player GetPlayer() {
		return player_;
	}

	/**
	 * Get Game Over flag.
	 * @return <code>true</code> if the game is over
	 */
	public boolean IsGameOver() {
		return gameOver_;
	}

	/**
	 * Get game-started flag.
	 * @return <code>true</code> if the game is started
	 */
	public boolean IsStartGame() {
		return startGame_;
	}

	/**
	 * Still in Ready->Go?
	 * @return <code>true</code> if the game is still in Ready->Go
	 */
	public boolean IsReadyTime() {
		return startGame_ && (curReadyTime_ < reqReadyTime_);
	}

	/**
	 * Did the Ready->Go end?
	 * @return <code>true</code> if the Ready->Go is ended and the game really started
	 */
	public boolean IsPlayTime() {
		return startGame_ && !IsReadyTime();
	}

	/**
	 * Is the game paused now?
	 * @return <code>true</code> if the game is currently paused
	 */
	public boolean IsPauseGame() {
		return startGame_ && pauseGame_;
	}

	/**
	 * Set button input flags.
	 * This is a convenience method. It just calls SetControllerFlags of the controller_ object.
	 * @param flags Button input flags
	 */
	public void SetControllerFlags(int flags) {
		controller_.SetControllerFlags(flags);
	}

	/**
	 * Get elapsed game time
	 * @return Elapsed game time
	 */
	public int GetPlayTime() {
		return playTime_;
	}

	/**
	 * Initialize
	 */
	public void Init() {
		playTime_ = 0;
		startGame_ = true;
		pauseGame_ = false;
		gameOver_ = false;
		statistics_.Init();
	}

	/**
	 * Begin the game
	 * @param randSeed Random-seed used by the randomizer
	 * @param delay Duration of Ready->Go status
	 */
	public void StartGame(long randSeed, int delay) {
		playfield_.Init(playerSetting_.PlayfieldWidth, playerSetting_.PlayfieldHeight);
		player_.Init(randSeed);
		controller_.Init();
		gameOver_ = false;
		curReadyTime_ = 0;
		reqReadyTime_ = delay;
		playTime_ = 0;
		startGame_ = true;
		pauseGame_ = false;
		statistics_.Init();
	}

	/**
	 * Quit the game
	 */
	public void EndGame() {
		startGame_ = false;

		for(int i = 0; i < eventReceivers_.size(); i++) {
			eventReceivers_.get(i).EndGame();
		}
	}

	/**
	 * Set the game-paused flag
	 * @param b <code>true</code> to pause, <code>false</code> to resume
	 */
	public void PauseGame(boolean b) {
		if(IsStartGame()) {
			pauseGame_ = b;

			for(int i = 0; i < eventReceivers_.size(); i++) {
				eventReceivers_.get(i).PauseGame(b);
			}
		}
	}

	/**
	 * Cause a Game Over
	 * @param reason Cause of death (Not really used)
	 */
	public void GameOver(int reason) {
		gameOver_ = true;

		for(int i = 0; i < eventReceivers_.size(); i++) {
			eventReceivers_.get(i).GameOver(reason);
		}
	}

	/**
	 * Run the main loop of the game
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void Run(int msec) {
		int maxExec = 17;
		remainLoopTime_ += msec;

		while(remainLoopTime_ > 0) {
			int loopTime = Math.min(remainLoopTime_, maxExec);
			msec = loopTime;
			remainLoopTime_ -= loopTime;

			if(IsStartGame() && !IsPauseGame()) {
				if(curReadyTime_ == 0) {
					for(int i = 0; i < eventReceivers_.size(); i++) {
						eventReceivers_.get(i).Ready(reqReadyTime_);
					}
				}
				if(curReadyTime_ < reqReadyTime_) {
					curReadyTime_ += msec;

					if(curReadyTime_ >= reqReadyTime_) {
						for(int i = 0; i < eventReceivers_.size(); i++) {
							eventReceivers_.get(i).Go();
						}
					} else {
						continue;
					}
				}

				playTime_ += msec;

				if(!IsGameOver()) {
					for(int i = 0; i < eventReceivers_.size(); i++) {
						eventReceivers_.get(i).OnUpdatePlayTime(playTime_);
					}

					controller_.Run(msec);

					if(controller_.IsBtnActivated(GarlicConst.BTN_SOFT)) player_.OnKeySoft(controller_.GetBtnActivatedCnt(GarlicConst.BTN_SOFT));
					if(controller_.IsBtnActivated(GarlicConst.BTN_HARD)) player_.OnKeyHard(controller_.GetBtnActivatedCnt(GarlicConst.BTN_HARD));
					if(controller_.IsBtnActivated(GarlicConst.BTN_HOLD)) player_.OnKeyHold(controller_.GetBtnActivatedCnt(GarlicConst.BTN_HOLD));
					if(controller_.IsBtnActivated(GarlicConst.BTN_LEFT)) player_.OnKeyLeft(controller_.GetBtnActivatedCnt(GarlicConst.BTN_LEFT));
					if(controller_.IsBtnActivated(GarlicConst.BTN_RIGHT)) player_.OnKeyRight(controller_.GetBtnActivatedCnt(GarlicConst.BTN_RIGHT));
					if(controller_.IsBtnActivated(GarlicConst.BTN_CW)) player_.OnKeyCW(controller_.GetBtnActivatedCnt(GarlicConst.BTN_CW));
					if(controller_.IsBtnActivated(GarlicConst.BTN_CCW)) player_.OnKeyCCW(controller_.GetBtnActivatedCnt(GarlicConst.BTN_CCW));

					for(int j = 0; j < GarlicConst.BTN_COUNT; j++) {
						if(controller_.IsBtnActivated(j)) {
							for(int i = 0; i < eventReceivers_.size(); i++) {
								eventReceivers_.get(i).ControllerBtnActivated(j, controller_.GetBtnActivatedCnt(j));
							}
						}
					}

					if(player_.GetCurPiece() != null) {
						player_.MovePiece(msec);
					}

					playfield_.Run(msec);

					if((player_.GetCurPiece() == null) && (!playfield_.IsBusy())) {
						player_.CreateNextPiece(msec);
					}
				}
			}
		}
	}

	/**
	 * Garbage Countering
	 * @param cnt Number of lines to reduce
	 * @return Remaining garbage lines
	 */
	public int ReduceAttack(int cnt) {
		for(int i = 0; i < attackEntries_.size(); i++) {
			if(cnt >= attackEntries_.get(i).cnt_) {
				cnt -= attackEntries_.get(i).cnt_;
				//attackEntries_.splice(i, i+1);
				for(int j = 0; j < i + 1; j++) {
					if(i < attackEntries_.size()) attackEntries_.remove(i);
				}
				i--;
			} else {
				attackEntries_.get(i).cnt_ -= cnt;
				cnt = 0;
			}
		}
		return cnt;
	}

	/**
	 * Add an attack entry
	 * @param fromId Specifies who sent this attack (?)
	 * @param cnt Number of garbage lines
	 * @param pattern Garbage block pattern (?)
	 */
	public void OnRecvAttack(int fromId, int cnt, int[] pattern) {
		if(cnt > 0) {
			AttackEntry entry = new AttackEntry();
			entry.fromId_ = fromId;
			entry.cnt_ = cnt;
			entry.pattern_ = pattern;
			attackEntries_.addElement(entry);
		}
	}

	/**
	 * Add garbage lines to the playfield
	 * @param cnt Number of garbage lines
	 * @param pattern Garbage block pattern (Unused)
	 */
	public void InsertLine(int cnt, int[] pattern) {
		int minY = playfield_.GetCurMinBlockY() - cnt;
		if(minY < 0) {
			GameOver(GarlicConst.GameOver_TopOut);
		}

		int[] tempArray = new int[playfield_.width_];
		for(int i = 0; i < playfield_.width_; i++) {
			tempArray[i] = GarlicConst.MinoColor_Garbage;
		}

		Random rand = new Random();
		tempArray[rand.nextInt(playfield_.width_)] = GarlicConst.MinoColor_Garbage;

		playfield_.InsertLine(cnt, tempArray, false);
	}

	/**
	 * Process the attack queue (It will add all garbage lines to the field)
	 */
	public void ProcessAttackQueue() {
		int cnt = 0;

		for(int i = 0; i < attackEntries_.size(); i++) {
			int cnt2 = (attackEntries_.get(i).cnt_ + cnt) / GetPlayfield().GetWidth();
			cnt += attackEntries_.get(i).cnt_ % GetPlayfield().GetWidth();
			if(cnt2 > 0) {
				InsertLine(cnt2, attackEntries_.get(i).pattern_);
			}
		}

		attackEntries_ = new Vector<AttackEntry>();

		if(cnt > 0) {
			OnRecvAttack(0, cnt, null);
		}
	}

	/**
	 * Get the total number of garbage lines in the queue
	 * @return Total number of garbage lines in the queue
	 */
	public int GetTotalAttackCnt() {
		int cnt = 0;
		for(int i = 0; i < attackEntries_.size(); i++) {
			cnt += attackEntries_.get(i).cnt_;
		}
		return cnt;
	}

	/**
	 * Attack information<br>
	 * [Anonymous class]
	 */
	public static class AttackEntry {
		/** Specifies who sent this attack (?) */
		public int fromId_ = 0;

		/** Number of garbage lines */
		public int cnt_ = 0;

		/** Garbage block pattern (?) */
		public int[] pattern_;
	}

	/**
	 * Score informations
	 */
	public static class Statistics {
		/** Line clear count. Not the number of lines cleared! ([0]=Single [1]=Double [2]=Triple [3]=Te*ris) */
		public int[] filledCount_;

		/** Number of garbage lines sent to the enemies */
		public int attack_ = 0;

		/** Number of pieces locked to the field */
		public int blockCnt_ = 0;

		/** Current Combo (REN) count */
		public int curComboCnt_ = 0;

		/** Max Combo (REN) count */
		public int maxComboCnt_ = 0;

		/** Back-to-Back chain count (2 or more to get x1.5 bonus) */
		public int b2bCnt_ = 0;

		/** Number of lines cleared */
		public int totalFilledLines_ = 0;

		/** T-Spin count */
		public int tspinCnt_ = 0;

		/** Number of lines cleared in the last piece */
		public int lastFilledLines_ = 0;

		/** <code>true</code> if there was a T-Spin in the last piece */
		public boolean lastTSpin_ = false;

		/** <code>true</code> if the last T-Spin was a mini-spin */
		public boolean lastCheapSpin_ = false;

		/** <code>true</code> if the last line-clear was a Perfect Clear (Bravo) */
		public boolean perfectClear_ = false;

		/** Number of garbage lines sent by normal line clears */
		public int tetrisStyleAttack_ = 0;

		/** Number of garbage lines sent by Combo (REN) */
		public int comboStyleAttack_ = 0;

		/** Number of garbage lines sent by T-Spin */
		public int tspinStyleAttack_ = 0;

		/** Level */
		public int curLevel_ = 1;

		/** Score */
		public int curScore_ = 0;

		/** Number of points earned in the last piece */
		public int thisGetScore_ = 0;

		/** Number of Hard Drop points earned in the last piece */
		public int thisHardDropScore_ = 0;

		/** Number of Soft Drop points earned in the last piece */
		public int thisSoftDropScore_ = 0;

		/** Number of Combo (REN) points earned in the last piece */
		public int thisComboScore_ = 0;

		/** Number of T-Spin points earned in the last piece */
		public int thisTSpinScore_ = 0;

		/** Number of line clear points earned in the last piece */
		public int thisFilledScore_ = 0;

		/** Number of Perfect Clear points earned in the last piece */
		public int thisPerfectClearScore_ = 0;

		/** Number of rows used with Soft Drop in the last piece */
		public int thisPieceSoftDropCnt_ = 0;

		/** Number of rows used with Hard Drop in the last piece */
		public int thisPieceHardDropCnt_ = 0;

		/**
		 * Initialize
		 */
		public void Init() {
			filledCount_ = new int[4];
			attack_ = 0;
			blockCnt_ = 0;
			curComboCnt_ = 0;
			maxComboCnt_ = 0;
			b2bCnt_ = 0;
			totalFilledLines_ = 0;
			tspinCnt_ = 0;
			lastFilledLines_ = 0;
			lastTSpin_ = false;
			lastCheapSpin_ = false;
			perfectClear_ = false;
			tetrisStyleAttack_ = 0;
			comboStyleAttack_ = 0;
			tspinStyleAttack_ = 0;
			curLevel_ = 1;
			curScore_ = 0;
			thisGetScore_ = 0;
			thisHardDropScore_ = 0;
			thisSoftDropScore_ = 0;
			thisComboScore_ = 0;
			thisTSpinScore_ = 0;
			thisFilledScore_ = 0;
			thisPerfectClearScore_ = 0;
			thisPieceSoftDropCnt_ = 0;
			thisPieceHardDropCnt_ = 0;
		}

		/**
		 * Is Back-to-Back bonus?
		 * @return <code>true</code> if there is a Back-to-Back bonus
		 */
		public boolean IsB2B() {
			return (b2bCnt_ >= 2) && (lastFilledLines_ >= 1);
		}

		/**
		 * Is T-Spin?
		 * @return <code>true</code> if there is a T-Spin bonus
		 */
		public boolean IsTSpin() {
			return lastTSpin_ || lastCheapSpin_;
		}
	}
}
