package tk.to_p.blankedoutbox.garlicclone.core;

/**
 * Player movements
 */
public class Player {
	/** GamePlay that owns this Player */
	public GamePlay gamePlay_ = null;

	/** Current piece */
	public Piece curPiece_ = new Piece();

	/** Timer for auto-fall */
	public int ellapseFallDownTime_ = 0;

	/** Timer for lock delay */
	public int ellapseLockDownTime_ = 0;

	/** Hold used flag, but for some reason this is a integer. Becomes 1 if Hold is used. Reset with 0 when the piece locked. */
	public int curHoldedCount_ = 0;

	/** Move/Rotate counter for lock delay reset limit */
	public int curExtendedPlacementCount_ = 0;

	/** Unused */
	public int curFallDownDelay_ = 0;

	/** Delay before the auto-fall timer begins (Not unlike PS2 Te*ris Kiwamemichi) */
	public int curFollowDownDelay_ = 0;

	/** Max Y position reached by the current piece (Used by lock delay reset) */
	public int curMaxPieceHeight_ = 0;

	/** Delay before piece appears (ARE) but it happens only when Hold is used. o_O */
	public int pieceWaitingTime_ = 0;

	/** Number of Holds used */
	public int totalHoldedCount_ = 0;

	/** Last wallkick level (0=No kicks used) */
	public int lastRotateLevel_ = 0;

	/** Hold piece ID */
	public int tetriMinoInHold_ = GarlicConst.TetriminoType_None;

	/** Piece sequence generator */
	public PlayerBag bag_ = new PlayerBag(this);

	/**
	 * Constructor
	 * @param gamePlay GamePlay that owns this Player
	 */
	public Player(GamePlay gamePlay) {
		gamePlay_ = gamePlay;
	}

	/**
	 * Initialize
	 * @param seed Random-Seed
	 */
	public void Init(long seed) {
		bag_.Init(seed);
		bag_.GetNext(0);
		ellapseFallDownTime_ = 0;
		ellapseLockDownTime_ = 0;
		lastRotateLevel_ = 0;
		curPiece_.type_ = GarlicConst.TetriminoType_None;
		curExtendedPlacementCount_ = 0;
		curFollowDownDelay_ = 0;
		pieceWaitingTime_ = 0;
		tetriMinoInHold_ = GarlicConst.TetriminoType_None;
		totalHoldedCount_ = 0;
	}

	/**
	 * Get current Tetrimino as a Piece object
	 * @return Current Tetrimino as a Piece object (Returns <code>null</code> if no piece is in action)
	 */
	public Piece GetCurPiece() {
		if(pieceWaitingTime_ > 0) {
			return null;
		}
		if(curPiece_.type_ == GarlicConst.TetriminoType_None) {
			return null;
		}
		return curPiece_;
	}

	/**
	 * Get the next piece (Does not remove)
	 * This is a convenience method. It just calls GetNext from the bag_ object.
	 * @param idx Piece Index
	 * @return Piece ID
	 * @throws ArrayIndexOutOfBoundsException If idx is less than 0 or bigger than the size of the bag
	 */
	public int GetTetriminoInBag(int idx) {
		return bag_.GetNext(idx);
	}

	/**
	 * Get hold piece ID
	 * @return Hold piece ID
	 */
	public int GetTetriminoInHold() {
		return tetriMinoInHold_;
	}

	/**
	 * Check blockout death (The newly generated piece overlaps to the existing blocks)
	 * @return <code>true</code> if the player gets a Game Over by a blockout
	 */
	public boolean CheckGameOver() {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;

			if(!gamePlay_.playfield_.TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				gamePlay_.GameOver(GarlicConst.GameOver_BlockOut);
				curPiece_.type_ = GarlicConst.TetriminoType_None;
				return true;
			}
		}
		return false;
	}

	/**
	 * Generate a new piece
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void CreateNextPiece(int msec) {
		if(pieceWaitingTime_ > 0) {
			pieceWaitingTime_ -= msec;
		}
		if((pieceWaitingTime_ > 0) || (GetCurPiece() != null)) {
			return;
		}

		gamePlay_.ProcessAttackQueue();

		ellapseLockDownTime_ = 0;
		ellapseFallDownTime_ = 0;
		curHoldedCount_ = 0;
		curExtendedPlacementCount_ = 0;
		curMaxPieceHeight_ = 0;
		lastRotateLevel_ = 0;

		curPiece_.SetPos(gamePlay_.GetPlayerSetting().PieceStartPosX, gamePlay_.GetPlayerSetting().PieceStartPosY);
		curPiece_.face_ = GarlicConst.FaceType_North;
		curPiece_.type_ = GetTetriminoInBag(0);
		curPiece_.id_++;
		bag_.Pop();

		gamePlay_.GetStatistics().thisPieceSoftDropCnt_ = 0;
		gamePlay_.GetStatistics().thisPieceHardDropCnt_ = 0;
		curFollowDownDelay_ = gamePlay_.GetPlayerSetting().FollowDownDelay;
		CalcGhostPos();

		for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
			gamePlay_.eventReceivers_.get(i).NewPiece(curPiece_);
		}

		CheckGameOver();
	}

	/**
	 * It will be called when the active piece has locked to the playfield.
	 * @param isTSpin <code>true</code> if there is a T-Spin
	 * @param isCheapSpin <code>true</code> if the T-Spin is a mini-spin.
	 */
	public void EndPiece(boolean isTSpin, boolean isCheapSpin) {
		int lines = gamePlay_.GetPlayfield().CheckAnimateLine();
		gamePlay_.GetStatistics().blockCnt_++;
		boolean isPerfectClear = false;

		if(!gamePlay_.GetPlayerSetting().CheapSpinRule && isCheapSpin) {
			isCheapSpin = false;
			isTSpin = false;
		}

		if(lines > 0) {
			if((lines == 4) || isTSpin || isCheapSpin) {
				gamePlay_.GetStatistics().b2bCnt_++;
			} else {
				gamePlay_.GetStatistics().b2bCnt_ = 0;
			}

			if(gamePlay_.GetPlayerSetting().ComboRule) {
				gamePlay_.GetStatistics().curComboCnt_++;
			}

			if(lines <= 4) {
				gamePlay_.GetStatistics().filledCount_[lines - 1]++;
			}

			isPerfectClear = (gamePlay_.GetPlayfield().GetCurMinBlockY() + lines == gamePlay_.GetPlayfield().GetPlayableHeight());

			GetCurAddRetObject retObject = new GetCurAddRetObject();
			int attack = GetCurAdd(gamePlay_, isTSpin, isCheapSpin, isPerfectClear, lines, retObject);
			gamePlay_.GetStatistics().attack_ += attack;
			gamePlay_.GetStatistics().tetrisStyleAttack_ += retObject.tetrisStyleAttack;
			gamePlay_.GetStatistics().comboStyleAttack_ += retObject.comboStyleAttack;
			gamePlay_.GetStatistics().tspinStyleAttack_ += retObject.tspinStyleAttack;

			for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
				gamePlay_.eventReceivers_.get(i).SendAttack(gamePlay_.ReduceAttack(attack));
			}
		} else {
			gamePlay_.GetStatistics().curComboCnt_ = 0;
		}

		gamePlay_.GetStatistics().maxComboCnt_ = Math.max(gamePlay_.GetStatistics().maxComboCnt_, gamePlay_.GetStatistics().curComboCnt_);
		if(isTSpin || isCheapSpin) gamePlay_.GetStatistics().tspinCnt_++;
		gamePlay_.GetStatistics().totalFilledLines_ += lines;
		gamePlay_.GetStatistics().lastFilledLines_ = lines;
		gamePlay_.GetStatistics().lastCheapSpin_ = (isCheapSpin && lines <= 1);
		gamePlay_.GetStatistics().lastTSpin_ = (!gamePlay_.GetStatistics().lastCheapSpin_ && isTSpin);
		gamePlay_.GetStatistics().perfectClear_ = isPerfectClear;

		GetCurScoreRetObject retObject2 = new GetCurScoreRetObject();
		int score = GetCurScore(gamePlay_, isTSpin, isCheapSpin, isPerfectClear, lines, gamePlay_.GetStatistics().curLevel_, retObject2);
		gamePlay_.GetStatistics().thisHardDropScore_ = retObject2.thisHardDropScore;
		gamePlay_.GetStatistics().thisSoftDropScore_ = retObject2.thisSoftDropScore;
		gamePlay_.GetStatistics().thisComboScore_ = retObject2.thisComboScore;
		gamePlay_.GetStatistics().thisTSpinScore_ = retObject2.thisTSpinScore;
		gamePlay_.GetStatistics().thisFilledScore_ = retObject2.thisFilledScore;
		gamePlay_.GetStatistics().thisPerfectClearScore_ = retObject2.thisPerfectClearScore;
		gamePlay_.GetStatistics().thisGetScore_ = score;
		gamePlay_.GetStatistics().curScore_ += score;

		for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
			gamePlay_.eventReceivers_.get(i).EndPiece(curPiece_);
		}

		curPiece_.type_ = GarlicConst.TetriminoType_None;
	}

	/**
	 * Downward movements of the active piece
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void MovePiece(int msec) {
		if(GetCurPiece() != null) {
			if(curFollowDownDelay_ > 0) {
				curFollowDownDelay_ -= msec;
			} else {
				ellapseFallDownTime_ += msec;
				ellapseLockDownTime_ += msec;

				int speed = gamePlay_.GetPlayerSetting().GetPieceFallSpeed(gamePlay_.GetStatistics().curLevel_);
				while(ellapseFallDownTime_ >= speed) {
					ellapseFallDownTime_ -= speed;
					OnFallDown(0, true);
				}
				if(ellapseLockDownTime_ >= gamePlay_.GetPlayerSetting().LockDownDelay) {
					OnLockDown();
				}
			}
		}
	}

	/**
	 * Lock the active piece to the playfield
	 */
	public void OnLockDown() {
		int x = curPiece_.x_;
		int y = curPiece_.y_;

		if(!gamePlay_.GetPlayfield().TestTetrimino(x, y + 1, curPiece_.GetTetriMino())) {
			Playfield.PutTetriminoInfos infos = new Playfield.PutTetriminoInfos();
			if(gamePlay_.GetPlayfield().PutTetrimino(x, y, false, curPiece_.GetTetriMino(), lastRotateLevel_, infos)) {
				EndPiece(infos.isTSpin, infos.isCheapSpin);
				if(infos.lockOut) {
					gamePlay_.GameOver(GarlicConst.GameOver_LockOut);
				}
			}
		}
	}

	/**
	 * It will be called when the piece falls
	 * @param cnt Number of rows
	 * @param bAuto <code>true</code> if the downward movement is caused by the native gravity. <code>false</code> if Soft or Hard drop is used.
	 */
	public void OnFallDown(int cnt, boolean bAuto) {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;
			y++;

			if(gamePlay_.GetPlayfield().TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				lastRotateLevel_ = 0;
				curPiece_.x_ = x;
				curPiece_.y_ = y;

				if(y > curMaxPieceHeight_) {
					curMaxPieceHeight_ = y;
					curExtendedPlacementCount_ = 0;
					ellapseLockDownTime_ = 0;
				}

				if(!bAuto) {
					gamePlay_.GetStatistics().thisPieceSoftDropCnt_++;

					for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
						gamePlay_.eventReceivers_.get(i).PieceSoft(curPiece_, cnt);
					}
				}

				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceFall(curPiece_, cnt);
				}
			}
		}
	}

	/**
	 * Move the active piece to the left
	 * @param cnt Button activated count
	 */
	public void OnKeyLeft(int cnt) {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;
			x--;

			if(gamePlay_.GetPlayfield().TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				if(gamePlay_.GetPlayerSetting().ExtendedPlacementCount < 0 ||
				   curExtendedPlacementCount_++ < gamePlay_.GetPlayerSetting().ExtendedPlacementCount)
				{
					ellapseLockDownTime_ = 0;
				}

				lastRotateLevel_ = 0;
				curPiece_.SetPos(x, y);
				CalcGhostPos();

				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceMoveLeft(curPiece_, cnt);
				}
			} else {
				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceMoveFail(curPiece_, cnt);
				}
			}
		}
	}

	/**
	 * Move the active piece to the right
	 * @param cnt Button activated count
	 */
	public void OnKeyRight(int cnt) {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;
			x++;

			if(gamePlay_.GetPlayfield().TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				if(gamePlay_.GetPlayerSetting().ExtendedPlacementCount < 0 ||
				   curExtendedPlacementCount_++ < gamePlay_.GetPlayerSetting().ExtendedPlacementCount)
				{
					ellapseLockDownTime_ = 0;
				}

				lastRotateLevel_ = 0;
				curPiece_.SetPos(x, y);
				CalcGhostPos();

				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceMoveRight(curPiece_, cnt);
				}
			} else {
				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceMoveFail(curPiece_, cnt);
				}
			}
		}
	}

	/**
	 * Rotate the active piece. Wallkicks are also here.
	 * @param tetriminos TetriMinos object (get it from GarlicConst)
	 * @param pos Position object that contains current X and Y positions
	 * @param face Current face
	 * @param newface New face
	 * @param level Wallkick level (0-5)
	 * @return 0=Failed, 1=Success, no kicks are used, 2 or bigger=Success, with wallkicks
	 */
	public int SuperRotate(TetriMinos tetriminos, Position pos, int face, int newface, int level) {
		/** Wallkick table for J,L,S,Z,T */
		int[][][] kickTable  = {{{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}},
							    {{0,0},{ 1,0},{ 1,1},{ 0,-2},{ 1,-2}},
							    {{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}},
							    {{0,0},{-1,0},{-1,1},{ 0,-2},{-1,-2}}};
		/** Wallkick table for I */
		int[][][] kickTableI = {{{0,0},{-1,0},{ 2,0},{-1, 0},{ 2, 0}},
							    {{0,0},{ 1,0},{ 1,0},{ 1,-1},{ 1, 2}},
							    {{0,0},{ 2,0},{-1,0},{ 2, 1},{-1, 1}},
							    {{0,0},{ 0,0},{ 0,0},{ 0, 2},{ 0,-1}}};
		/** Wallkick table for O (is this really needed?) */
		int[][][] kickTableO = {{{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}},
							    {{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}},
							    {{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}},
							    {{0,0},{ 0,0},{ 0,0},{ 0, 0},{ 0, 0}}};

		for(int i = 0; i < level; i++) {
			int x1, y1, x2, y2, x3, y3;

			switch(tetriminos.type_) {
			case GarlicConst.TetriminoType_O:
				x1 = kickTableO[face][i][0];
				y1 = kickTableO[face][i][1];
				x2 = kickTableO[newface][i][0];
				y2 = kickTableO[newface][i][1];
				break;
			case GarlicConst.TetriminoType_I:
				x1 = kickTableI[face][i][0];
				y1 = kickTableI[face][i][1];
				x2 = kickTableI[newface][i][0];
				y2 = kickTableI[newface][i][1];
				break;
			default:
				x1 = kickTable[face][i][0];
				y1 = kickTable[face][i][1];
				x2 = kickTable[newface][i][0];
				y2 = kickTable[newface][i][1];
				break;
			}
			x3 = x1 - x2;
			y3 = y1 - y2;

			if (gamePlay_.playfield_.TestTetrimino(pos.x + x3, pos.y + y3, tetriminos.tetriMinos_[newface])) {
				pos.x += x3;
				pos.y += y3;
				return i + 1;
			}
		}

		return 0;
	}

	/**
	 * Clockwise Rotation
	 * @param cnt Button activated count
	 */
	public void OnKeyCW(int cnt) {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;
			int newface = curPiece_.face_ + 1;
			if(newface > GarlicConst.FaceType_West) newface = GarlicConst.FaceType_North;

			Position pos = new Position(x, y);
			int rotatelevel = SuperRotate(GarlicConst.TetriMinos_[curPiece_.type_], pos, curPiece_.GetFace(), newface,
										  gamePlay_.GetPlayerSetting().RotationLevel);

			if(rotatelevel > 0) {
				if(gamePlay_.GetPlayerSetting().ExtendedPlacementCount < 0 ||
				   curExtendedPlacementCount_++ < gamePlay_.GetPlayerSetting().ExtendedPlacementCount)
				{
					ellapseLockDownTime_ = 0;
				}

				lastRotateLevel_ = rotatelevel;
				curPiece_.SetPos(pos.x, pos.y);
				curPiece_.SetFace(newface);
				CalcGhostPos();

				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceRotateCW(curPiece_, cnt);
				}
			} else {
				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceRotateFail(curPiece_, cnt);
				}
			}
		}
	}

	/**
	 * Counter-Clockwise Rotation
	 * @param cnt Button activated count
	 */
	public void OnKeyCCW(int cnt) {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;
			int newface = curPiece_.face_ - 1;
			if(newface < GarlicConst.FaceType_North) newface = GarlicConst.FaceType_West;

			Position pos = new Position(x, y);
			int rotatelevel = SuperRotate(GarlicConst.TetriMinos_[curPiece_.type_], pos, curPiece_.GetFace(), newface,
										  gamePlay_.GetPlayerSetting().RotationLevel);

			if(rotatelevel > 0) {
				if(gamePlay_.GetPlayerSetting().ExtendedPlacementCount < 0 ||
				   curExtendedPlacementCount_++ < gamePlay_.GetPlayerSetting().ExtendedPlacementCount)
				{
					ellapseLockDownTime_ = 0;
				}

				lastRotateLevel_ = rotatelevel;
				curPiece_.SetPos(pos.x, pos.y);
				curPiece_.SetFace(newface);
				CalcGhostPos();

				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceRotateCCW(curPiece_, cnt);
				}
			} else {
				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceRotateFail(curPiece_, cnt);
				}
			}
		}
	}

	/**
	 * Soft Drop
	 * @param cnt Button activated count
	 */
	public void OnKeySoft(int cnt) {
		if(GetCurPiece() != null) {
			curFollowDownDelay_ = 0;
			OnFallDown(cnt, false);
		}
	}

	/**
	 * Hard Drop
	 * @param cnt Button activated count
	 */
	public void OnKeyHard(int cnt) {
		if((GetCurPiece() != null) && (gamePlay_.GetPlayerSetting().EnableHardDrop)) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;

			Playfield.PutTetriminoInfos infos = new Playfield.PutTetriminoInfos();
			if(gamePlay_.GetPlayfield().PutTetrimino(x, y, true, curPiece_.GetTetriMino(), lastRotateLevel_, infos)) {
				for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
					gamePlay_.eventReceivers_.get(i).PieceHard(curPiece_, infos.downLineCnt, cnt);
				}

				curPiece_.SetPos(x, y + infos.downLineCnt);
				gamePlay_.GetStatistics().thisPieceHardDropCnt_ = infos.downLineCnt;
				EndPiece(infos.isTSpin, infos.isCheapSpin);

				if(infos.lockOut) {
					gamePlay_.GameOver(GarlicConst.GameOver_LockOut);
				}
			}
		}
	}

	/**
	 * Can Hold be used now?
	 * @return <code>true</code> if Hold can be used now
	 */
	public boolean CanHold() {
		return (GetCurPiece() != null) && (curHoldedCount_ == 0) && (GetRemainHoldCount() != 0);
	}

	/**
	 * Get number of Holds remaining
	 * @return Number of Holds remaining (-1=Unlimited Holds Available)
	 */
	public int GetRemainHoldCount() {
		if(gamePlay_.GetPlayerSetting().HoldBlockCount < 0) {
			return -1;
		}
		return gamePlay_.GetPlayerSetting().HoldBlockCount - totalHoldedCount_;
	}

	/**
	 * Hold the active piece if possible
	 * @param cnt Button activated count
	 */
	public void OnKeyHold(int cnt) {
		if(CanHold()) {
			ellapseFallDownTime_ = 0;
			ellapseLockDownTime_ = 0;
			curExtendedPlacementCount_ = 0;
			curMaxPieceHeight_ = 0;
			lastRotateLevel_ = 0;

			int type = curPiece_.type_;

			if(tetriMinoInHold_ != GarlicConst.TetriminoType_None) {
				curPiece_.type_ = tetriMinoInHold_;
			} else {
				curPiece_.type_ = GetTetriminoInBag(0);
				bag_.Pop();
			}

			curHoldedCount_++;
			totalHoldedCount_++;

			tetriMinoInHold_ = type;

			curPiece_.SetPos(gamePlay_.GetPlayerSetting().PieceStartPosX, gamePlay_.GetPlayerSetting().PieceStartPosY);
			curPiece_.face_ = GarlicConst.FaceType_North;
			curPiece_.id_++;
			CalcGhostPos();

			for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
				gamePlay_.eventReceivers_.get(i).PieceHold(curPiece_, cnt);
			}

			CheckGameOver();
			pieceWaitingTime_ = gamePlay_.GetPlayerSetting().HoldDelay;
		} else {
			for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
				gamePlay_.eventReceivers_.get(i).PieceHoldFail(curPiece_, cnt);
			}
		}
	}

	/**
	 * Update the ghost piece position
	 */
	public void CalcGhostPos() {
		if(GetCurPiece() != null) {
			int x = curPiece_.x_;
			int y = curPiece_.y_;

			while(gamePlay_.GetPlayfield().TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				y++;
			}
			y--;

			if(gamePlay_.GetPlayfield().TestTetrimino(x, y, curPiece_.GetTetriMino())) {
				curPiece_.xGhost_ = x;
				curPiece_.yGhost_ = y;
			}
		}
	}

	/**
	 * Get the attack points
	 * @param gamePlay GamePlay object
	 * @param isTSpin <code>true</code> if T-Spin
	 * @param isCheapSpin <code>true</code> if mini-spin
	 * @param perfectClear <code>true</code> if Perfect Clear
	 * @param filledLine Number of lines cleared
	 * @param retObject The detailed result will be stored here
	 * @return Attack points (Number of garbage lines sent * 10)
	 */
	public static int GetCurAdd(GamePlay gamePlay, boolean isTSpin, boolean isCheapSpin, boolean perfectClear, int filledLine,
								GetCurAddRetObject retObject)
	{
		retObject.tetrisStyleAttack = 0;
		retObject.comboStyleAttack = 0;
		retObject.tspinStyleAttack = 0;

		boolean isB2B = gamePlay.GetStatistics().IsB2B();

		if(isTSpin) {
			if(filledLine == 0) {
				retObject.tspinStyleAttack += gamePlay.GetPlayerSetting().AttackTSpin0;
			} else if(filledLine == 1) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin1 : gamePlay.GetPlayerSetting().AttackTSpin1;
			} else if(filledLine == 2) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin2 : gamePlay.GetPlayerSetting().AttackTSpin2;
			} else if(filledLine == 3) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin3 : gamePlay.GetPlayerSetting().AttackTSpin3;
			}
		} else if(isCheapSpin) {
			if(filledLine == 0) {
				retObject.tspinStyleAttack += gamePlay.GetPlayerSetting().AttackTSpin0;
			} else if(filledLine == 1) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin1Mini : gamePlay.GetPlayerSetting().AttackTSpin1Mini;
			} else if(filledLine == 2) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin2 : gamePlay.GetPlayerSetting().AttackTSpin2;
			} else if(filledLine == 3) {
				retObject.tspinStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTSpin3 : gamePlay.GetPlayerSetting().AttackTSpin3;
			}
		} else {
			if(filledLine == 1) {
				retObject.tetrisStyleAttack += gamePlay.GetPlayerSetting().AttackSingle;
			} else if(filledLine == 2) {
				retObject.tetrisStyleAttack += gamePlay.GetPlayerSetting().AttackDouble;
			} else if(filledLine == 3) {
				retObject.tetrisStyleAttack += gamePlay.GetPlayerSetting().AttackTriple;
			} else if(filledLine == 4) {
				retObject.tetrisStyleAttack += isB2B ? gamePlay.GetPlayerSetting().AttackB2BTetris : gamePlay.GetPlayerSetting().AttackTetris;
			}
		}

		if (perfectClear) {
			retObject.tetrisStyleAttack += gamePlay.GetPlayerSetting().AttackPerfectClear;
		}

		int comboStart = gamePlay.GetPlayerSetting().AttackComboStartCount;
		if (gamePlay.GetStatistics().curComboCnt_ >= comboStart) {
			int comboCount = gamePlay.GetStatistics().curComboCnt_ - comboStart;
			if (comboCount >= gamePlay.GetPlayerSetting().AttackCombos.length) {
				comboCount = gamePlay.GetPlayerSetting().AttackCombos.length - 1;
			}
			int comboAttack = gamePlay.GetPlayerSetting().AttackCombos[comboCount];
			retObject.comboStyleAttack += comboAttack;
		}

		return retObject.tetrisStyleAttack + retObject.comboStyleAttack + retObject.tspinStyleAttack;
	}

	/**
	 * Get the points earned
	 * @param gamePlay GamePlay object
	 * @param isTSpin <code>true</code> if T-Spin
	 * @param isCheapSpin <code>true</code> if mini-spin
	 * @param perfectClear <code>true</code> if Perfect Clear
	 * @param filledLine Number of lines cleared
	 * @param curLevel Current level (Starts from 1, not 0!)
	 * @param retObject The detailed result will be stored here
	 * @return Points
	 */
	public static int GetCurScore(GamePlay gamePlay, boolean isTSpin, boolean isCheapSpin, boolean perfectClear, int filledLine,
								  int curLevel, GetCurScoreRetObject retObject)
	{
		boolean isB2B = gamePlay.GetStatistics().IsB2B();

		retObject.thisHardDropScore = 0;
		retObject.thisComboScore = 0;
		retObject.thisTSpinScore = 0;
		retObject.thisFilledScore = 0;
		retObject.thisPerfectClearScore = 0;
		retObject.thisSoftDropScore = gamePlay.GetStatistics().thisPieceSoftDropCnt_;
		retObject.thisHardDropScore += gamePlay.GetStatistics().thisPieceHardDropCnt_ * 2;

		if(isCheapSpin) {
			if(filledLine == 0) {
				retObject.thisTSpinScore += gamePlay.GetPlayerSetting().ScoreTSpin0;
			} else if(filledLine == 1) {
				retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin1Mini : gamePlay.GetPlayerSetting().ScoreTSpin1Mini;
			} else if(filledLine == 2) {
				retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin2 : gamePlay.GetPlayerSetting().ScoreTSpin2;
			} else if(filledLine == 3) {
				retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin3 : gamePlay.GetPlayerSetting().ScoreTSpin3;
			}
		} else {
			if(isTSpin) {
				if(filledLine == 0) {
					retObject.thisTSpinScore += gamePlay.GetPlayerSetting().ScoreTSpin0;
				} else if(filledLine == 1) {
					retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin1 : gamePlay.GetPlayerSetting().ScoreTSpin1;
				} else if(filledLine == 2) {
					retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin2 : gamePlay.GetPlayerSetting().ScoreTSpin2;
				} else if(filledLine == 3) {
					retObject.thisTSpinScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTSpin3 : gamePlay.GetPlayerSetting().ScoreTSpin3;
				}
			} else {
				if(filledLine == 1) {
					retObject.thisFilledScore += gamePlay.GetPlayerSetting().ScoreSingle;
				} else if(filledLine == 2) {
					retObject.thisFilledScore += gamePlay.GetPlayerSetting().ScoreDouble;
				} else if(filledLine == 3) {
					retObject.thisFilledScore += gamePlay.GetPlayerSetting().ScoreTriple;
				} else if(filledLine == 4) {
					retObject.thisFilledScore += isB2B ? gamePlay.GetPlayerSetting().ScoreB2BTetris : gamePlay.GetPlayerSetting().ScoreTetris;
				}
			}
		}

		if(perfectClear) {
			retObject.thisPerfectClearScore += gamePlay.GetPlayerSetting().ScorePerfectClear;
		}

		if(gamePlay.GetStatistics().curComboCnt_ >= gamePlay.GetPlayerSetting().ScoreComboStartCount) {
			GamePlay.Statistics stats = gamePlay.GetStatistics();
			PlayerSetting setting = gamePlay.GetPlayerSetting();
			retObject.thisComboScore += ((stats.curComboCnt_ - setting.ScoreComboStartCount) / setting.ScoreComboInterval + 1) * setting.ScoreComboSize;
		}

		retObject.thisComboScore *= curLevel;
		retObject.thisTSpinScore *= curLevel;
		retObject.thisFilledScore *= curLevel;
		retObject.thisPerfectClearScore *= curLevel;

		return  retObject.thisHardDropScore +
				retObject.thisComboScore +
				retObject.thisTSpinScore +
				retObject.thisFilledScore +
				retObject.thisPerfectClearScore +
				retObject.thisSoftDropScore;
	}

	/**
	 * Current Tetrimino Data
	 */
	public static class Piece {
		/** Current X position */
		public int x_ = 0;

		/** Current Y position */
		public int y_ = 0;

		/** X position of ghost */
		public int xGhost_ = 0;

		/** Y position of ghost */
		public int yGhost_ = 0;

		/** Face */
		public int face_ = GarlicConst.FaceType_North;

		/** Tetrimino Type */
		public int type_ = GarlicConst.TetriminoType_None;

		/** ID (Starts from 0. Increases when the new piece gets generated.) */
		public int id_ = 0;

		/**
		 * Get TetriMino object
		 * @return TetriMino object
		 */
		public TetriMinos.TetriMino GetTetriMino() {
			return GarlicConst.TetriMinos_[type_].tetriMinos_[face_];
		}

		/**
		 * Set the current position
		 * @param x X position
		 * @param y Y position
		 */
		public void SetPos(int x, int y) {
			x_ = x;
			y_ = y;
		}

		/**
		 * Set the face
		 * @param face Face
		 */
		public void SetFace(int face) {
			face_ = face;
		}

		/**
		 * Get the face
		 * @return Face
		 */
		public int GetFace() {
			return face_;
		}
	}

	/**
	 * Detailed results of GetCurAdd<br>
	 * [Anonymous class]
	 */
	public static class GetCurAddRetObject {
		/** Attacks from regular line clears + perfect clears */
		public int tetrisStyleAttack = 0;

		/** Attacks from Combo */
		public int comboStyleAttack = 0;

		/** Attacks from T-Spin line clears */
		public int tspinStyleAttack = 0;
	}

	/**
	 * Detailed results of GetCurScore<br>
	 * [Anonymous class]
	 */
	public static class GetCurScoreRetObject {
		/** Points from regular line clears */
		public int thisFilledScore = 0;

		/** Points from Combo */
		public int thisComboScore = 0;

		/** Points from T-Spin */
		public int thisTSpinScore = 0;

		/** Points from Perfect Clear */
		public int thisPerfectClearScore = 0;

		/** Points from Soft Drop */
		public int thisSoftDropScore = 0;

		/** Points from Hard Drop */
		public int thisHardDropScore = 0;
	}
}
