package tk.to_p.blankedoutbox.garlicclone.core;
import java.util.Vector;

/**
 * Playfield
 */
public class Playfield {
	/** GamePlay that owns this Playfield */
	public GamePlay gamePlay_ = null;

	/** Width */
	public int width_ = 0;

	/** Height */
	public int height_ = 0;

	/** Grids */
	public Grid[] grids_ = null;

	/** Minimum Y (stack height) */
	public int curMinBlockY_ = 0;

	/** Number of concrete lines */
	public int curConcreteLineCnt_ = 0;

	/** Elapsed time */
	public int curTime_ = 0;

	/** Animation status of the disappearing lines */
	public AnimatingLine animatingLine_ = new AnimatingLine();

	/** Falling status of the stack */
	public FallingLine fallingLine_ = new FallingLine();

	/**
	 * Constructor
	 * @param gamePlay GamePlay that owns this Playfield
	 */
	public Playfield(GamePlay gamePlay) {
		gamePlay_ = gamePlay;
	}

	/**
	 * Get field width
	 * @return Width
	 */
	public int GetWidth() {
		return width_;
	}

	/**
	 * Get field height
	 * @return Height
	 */
	public int GetHeight() {
		return height_;
	}

	/**
	 * Get stack fall speed
	 * @param idx Y position
	 * @return Stack fall speed
	 */
	public float GetLineAddY(int idx) {
		if(IsLineFalling()) {
			if(idx < fallingLine_.vecIdx_.get(0)) {
				return fallingLine_.ratio_;
			}
		}
		return 0f;
	}

	/**
	 * Get elapsed time
	 * @return Elapsed time
	 */
	public int GetCurTime() {
		return curTime_;
	}

	/**
	 * Clear all blocks in the field
	 */
	public void Clear() {
		grids_ = new Grid[width_ * height_];

		for(int i = 0; i < width_ * height_; i++) {
			grids_[i] = new Grid();
			grids_[i].mino_ = new Mino();
			grids_[i].empty_ = true;
			grids_[i].createTime_ = 0;
		}
	}

	/**
	 * Initialize
	 * @param width Width
	 * @param height Height
	 */
	public void Init(int width, int height) {
		width_ = width;
		height_ = height;
		curMinBlockY_ = height;
		curTime_ = 0;
		Clear();
	}

	/**
	 * Get specfic Grid
	 * @param x X position
	 * @param y Y position
	 * @return Grid data
	 */
	public Grid GetGrid(int x, int y) {
		return grids_[x + y * width_];
	}

	/**
	 * Get the number of concrete lines
	 * @return Number of concrete lines
	 */
	public int GetConcreteLineCnt() {
		return curConcreteLineCnt_;
	}

	/**
	 * Get the playable height
	 * @return Playable height (Field height - Number of concrete lines)
	 */
	public int GetPlayableHeight() {
		return height_ - GetConcreteLineCnt();
	}

	/**
	 * Is the specific Grid filled with something?
	 * @param x X position
	 * @param y Y position
	 * @return <code>true</code> if the specified Grid is filled with something, or when the specified position is out-of-bounds.
	 */
	public boolean IsGridFilled(int x, int y) {
		return (x < 0) || (y >= height_) || (x >= width_) || (y < 0) || !GetGrid(x, y).empty_;
	}

	/**
	 * Is the specific Grid filled with a block?
	 * @param x X position
	 * @param y Y position
	 * @return <code>true</code> if the specified Grid is filled with a block.
	 *         Unlike IsGridFilled, it returns <code>false</code> if the specified position is out-of-bounds.
	 */
	public boolean IsMinoExist(int x, int y) {
		return (x >= 0) && (x < width_) && (y >= 0) && (y < height_) && !GetGrid(x, y).empty_;
	}

	/**
	 * Is the specific Grid filled with a specific colored block?
	 * @param x X position
	 * @param y Y position
	 * @param clr Color
	 * @return <code>true</code> if the specified Grid is filled with a specified-colored block
	 */
	public boolean IsMinoColor(int x, int y, int clr) {
		 return IsMinoExist(x, y) && (GetGrid(x, y).mino_.color_ == clr);
	}

	/**
	 * Set a 1-cell block to the field
	 * @param x X position
	 * @param y Y position
	 * @param mino Mino (a 1-cell block)
	 * @return The new Grid
	 */
	public Grid SetMino(int x, int y, Mino mino) {
		curMinBlockY_ = Math.min(curMinBlockY_, y);
		if(mino.color_ == GarlicConst.MinoColor_Concrete) {
			curConcreteLineCnt_ = Math.max(curConcreteLineCnt_, GetHeight() - y);
		}
		Grid grid = GetGrid(x, y);
		grid.mino_.color_ = mino.color_;
		grid.mino_.sticky_ = mino.sticky_;
		grid.empty_ = (mino.color_ == GarlicConst.MinoColor_None);
		return grid;
	}

	/**
	 * Test the Tetrimino collision
	 * @param x X position
	 * @param y Y position
	 * @param tetriMino Tetrimino to test the collision
	 * @return Returns <code>true</code> if the Tetrimino does <b>NOT</b> overlaps to the existing blocks.
	 *         In other words, it returns <code>true</code> if the Tetrimino can be moved to the specified position.
	 */
	public boolean TestTetrimino(int x, int y, TetriMinos.TetriMino tetriMino) {
		for(int i = 0; i < GarlicConst.Tetri; i++) {
			int dx = tetriMino.minos_[i].x_;
			int dy = tetriMino.minos_[i].y_;

			if(!((x + dx < width_) && (y + dy < height_) && (x + dx >= 0) && (y + dy >= 0))) {
				return false;
			} else if(!GetGrid(x + dx, y + dy).empty_) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Put a Tetrimino to the field
	 * @param x X position
	 * @param y Y position
	 * @param drop <code>true</code> to drop the Tetrimino to the surface
	 * @param tetriMino Tetrimino object
	 * @param rotateLevel Last wallkick level (0-5)
	 * @param infos The detailed result will be stored here
	 * @return <code>true</code> if success
	 */
	public boolean PutTetrimino(int x, int y, boolean drop, TetriMinos.TetriMino tetriMino, int rotateLevel, PutTetriminoInfos infos) {
		int y2 = y;
		infos.downLineCnt = 0;

		if(drop) {
			while(TestTetrimino(x, y, tetriMino)) {
				y++;
			}
			y--;
		}

		if(TestTetrimino(x, y, tetriMino)) {
			infos.downLineCnt = y - y2;
			infos.isTSpin = false;
			infos.isCheapSpin = false;

			// T-Spin check
			if( (rotateLevel > 0) && (infos.downLineCnt == 0) && (tetriMino.type_ == GarlicConst.TetriminoType_T) &&
				(gamePlay_.GetPlayerSetting().tSpinRule) )
			{
				int corners = 0;
				if(x < 0 || !GetGrid(x, y).empty_) {
					corners++;
				}
				if(x + 2 >= width_ || !GetGrid(x + 2, y).empty_) {
					corners++;
				}
				if(x < 0 || y + 2 >= height_ || !GetGrid(x, y + 2).empty_) {
					corners++;
				}
				if(x + 2 >= width_ || y + 2 >= height_ || !GetGrid(x + 2, y + 2).empty_) {
					corners++;
				}
				infos.isTSpin = (corners >= 3);

				// T-Spin Mini check. If you remove the rotateLevel>1 it will be same as Windows version of TOJ.
				if(infos.isTSpin &&
				   ( rotateLevel>1 ||
				    (GarlicConst.FaceType_North == tetriMino.face_ && IsGridFilled(x+1, y+2) ||
				     GarlicConst.FaceType_East  == tetriMino.face_ && IsGridFilled(x  , y+1) ||
				     GarlicConst.FaceType_South == tetriMino.face_ && IsGridFilled(x+1, y  ) ||
				     GarlicConst.FaceType_West  == tetriMino.face_ && IsGridFilled(x+2, y+1)) )
				  )
				{
					infos.isCheapSpin = true;
				}
			}

			boolean notLockout = false;
			for(int i = 0; i < GarlicConst.Tetri; i++) {
				int dx = tetriMino.minos_[i].x_;
				int dy = tetriMino.minos_[i].y_;
				Grid grid = SetMino(x + dx, y + dy, tetriMino.minos_[i]);

				grid.createTime_ = curTime_;

				if(!notLockout && (y + dy >= height_ / 2)) {
					notLockout = true;
				}
			}
			infos.lockOut = !notLockout;

			return true;
		}

		return false;
	}

	/**
	 * Something to do with line-clear animations
	 * @return Number of lines to animate
	 */
	public int CheckAnimateLine() {
		Vector<Integer> arIdx = new Vector<Integer>();
		int lines = 0;

		for(int y = height_ - 1; y >= curMinBlockY_; y--) {
			boolean notEmpty = true;

			for(int x = 0; x < width_; x++) {
				Grid grid = grids_[x + y * width_];
				if(notEmpty && grid.empty_) {
					notEmpty = false;
				}
			}

			if(notEmpty) {
				boolean isAnimate = false;

				for(int x = 0; x < width_; x++) {
					Grid grid = grids_[x + y * width_];
					if(grid.mino_.color_ != GarlicConst.MinoColor_Concrete) {
						isAnimate = true;
						break;
					}
				}

				if(isAnimate) {
					arIdx.addElement(y);
					lines++;
				}
			}
		}

		if(!arIdx.isEmpty()) {
			AnimateLine(arIdx, gamePlay_.GetPlayerSetting().LineLifeTime);
		}

		return lines;
	}

	/**
	 * Something to do with line-clear animations
	 * @param arIdx Dynamic array that contains Y position of cleared lines?
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void AnimateLine(Vector<Integer> arIdx, int msec) {
		for(int i = 0; i < arIdx.size(); i++) {
			animatingLine_.vecIdx_.addElement(arIdx.get(i));
			animatingLine_.lifeLimitMSec_ = msec;
		}

		for(int i = 0; i < gamePlay_.eventReceivers_.size(); i++) {
			gamePlay_.eventReceivers_.get(i).AnimateLine(arIdx, msec);
		}
	}

	/**
	 * Clear a Mino
	 * @param x X position
	 * @param y Y position
	 */
	public void ClearMino(int x, int y) {
		grids_[x + y * width_].empty_ = true;
	}

	/**
	 * Line clear
	 * @param idx Y position
	 */
	public void ClearLine(int idx) {
		if(idx == curMinBlockY_) {
			curMinBlockY_++;
		}

		for(int i = 0; i < width_; i++) {
			grids_[i + idx * width_].empty_ = true;
		}

		for(int i = 0; i < width_; i++) {
			if(idx > 0) {
				Mino mino = grids_[i + (idx-1) * width_].mino_;
				int sticky = mino.sticky_;
				sticky &= ~GarlicConst.StickyFlag_S;
				mino.sticky_ = sticky;
			}
			if(idx < height_ - 1) {
				Mino mino = grids_[i + (idx+1) * width_].mino_;
				int sticky = mino.sticky_;
				sticky &= ~GarlicConst.StickyFlag_N;
				mino.sticky_ = sticky;
			}
		}
	}

	/**
	 * Insert garbage line(s)
	 * @param cnt Number of lines to add
	 * @param colors Block colors
	 * @param concrete Unused
	 */
	public void InsertLine(int cnt, int[] colors, boolean concrete) {
		if(cnt > GetHeight()) cnt = GetHeight();
		curMinBlockY_ -= cnt;

		int concreteCount = curConcreteLineCnt_;

		for(int i = cnt; i < height_ - concreteCount; i++) {
			for(int j = 0; j < width_; j++) {
				Grid gridDest = grids_[j + (i - cnt) * width_];
				Grid gridSrc = grids_[j + i * width_];
				gridDest.mino_.color_ = gridSrc.mino_.color_;
				gridDest.mino_.sticky_ = gridSrc.mino_.sticky_;
				gridDest.empty_ = gridSrc.empty_;
				gridDest.createTime_ = gridSrc.createTime_;
			}
		}

		int offsetY = height_ - cnt - concreteCount;
		for(int i = 0; i < cnt; i++) {
			for(int j = 0; j < width_; j++) {
				SetMino(j, offsetY + i, new Mino(colors[j + i * width_]));
			}
		}

		int offsetConcrete = height_ - concreteCount;
		for(int i = offsetConcrete - cnt; i < offsetConcrete; i++) {
			for(int j = 0; j < width_; j++) {
				CalcMinoSticky(j, i, i != offsetConcrete-1, i != offsetConcrete-cnt);
			}
		}
	}

	/**
	 * Set connection flags
	 * @param x X position
	 * @param y Y position
	 * @param top Upward check
	 * @param bottom Downward check
	 */
	public void CalcMinoSticky(int x, int y, boolean top, boolean bottom) {
		int sticky = 0;
		int color = GetGrid(x, y).mino_.color_;

		if(bottom) {
			if(IsMinoColor(x, y - 1, color)) {
				sticky |= GarlicConst.StickyFlag_N;
			}
		}
		if(IsMinoColor(x + 1, y, color)) {
			sticky |= GarlicConst.StickyFlag_E;
		}
		if(IsMinoColor(x - 1, y, color)) {
			sticky |= GarlicConst.StickyFlag_W;
		}
		if(top) {
			if(IsMinoColor(x, y + 1, color)) {
				sticky |= GarlicConst.StickyFlag_S;
			}
		}

		GetGrid(x, y).mino_.sticky_ = sticky;
	}

	/**
	 * Something to do with line-clear
	 * @param idx Y position
	 */
	public void RemoveLine(int idx) {
		if(idx >= curMinBlockY_) {
			curMinBlockY_++;
		}

		for(int i = idx - 1; i >= 0; i--) {
			for(int j = 0; j < width_; j++) {
				Grid gridDest = grids_[j + (i + 1) * width_];
				Grid gridSrc = grids_[j + i * width_];
				gridDest.mino_.color_ = gridSrc.mino_.color_;
				gridDest.mino_.sticky_ = gridSrc.mino_.sticky_;
				gridDest.empty_ = gridSrc.empty_;
				gridDest.createTime_ = gridSrc.createTime_;
			}
		}

		ClearLine(0);
	}

	/**
	 * Freefall? Unused from the core engine.
	 */
	public void DropAllEmptyMino() {
		for(int j = 0; j < width_; j++) {
			for(int i = GetHeight() - 1; i >= 0; i--) {
				if(!IsMinoExist(j, i)) {
					for(int k = i - 1; k >= 0; k--) {
						if(IsMinoExist(j, k)) {
							Grid gridDest = grids_[j + i * width_];
							Grid gridSrc = grids_[j + k * width_];
							gridDest.mino_.color_ = gridSrc.mino_.color_;
							gridDest.mino_.sticky_ = gridSrc.mino_.sticky_;
							gridDest.empty_ = gridSrc.empty_;
							gridDest.createTime_ = gridSrc.createTime_;
							gridSrc.empty_ = true;
							break;
						}
					}
				}
			}
		}

		for(int j = 0; j < width_; j++) {
			if(!IsMinoExist(j, GetHeight() - 1)) {
				for(int i = j + 1; i < width_; i++) {
					if(IsMinoExist(i, GetHeight()-1)) {
						for(int k = GetHeight()-1; k >= 0; k--) {
							Grid gridDest = grids_[j + k * width_];
							Grid gridSrc = grids_[i + k * width_];
							gridDest.mino_.color_ = gridSrc.mino_.color_;
							gridDest.mino_.sticky_ = gridSrc.mino_.sticky_;
							gridDest.empty_ = gridSrc.empty_;
							gridDest.createTime_ = gridSrc.createTime_;
							gridSrc.empty_ = true;
						}
						break;
					}
				}
			}
		}
	}

	/**
	 * Is in line-clearing sequence?
	 * @return <code>true</code> if line-clearing sequence
	 */
	public boolean IsBusy() {
		return IsLineFalling() || IsLineAnimating();
	}

	/**
	 * Is the stack falling?
	 * @return <code>true</code> if the stack is falling
	 */
	public boolean IsLineFalling() {
		return !fallingLine_.vecIdx_.isEmpty();
	}

	/**
	 * Are the cleared-lines animating?
	 * @return <code>true</code> if the cleared-lines are animating
	 */
	public boolean IsLineAnimating() {
		return !animatingLine_.vecIdx_.isEmpty();
	}

	/**
	 * Get the minimum Y position (stack height)
	 * @return Minimum Y position (stack height)
	 */
	public int GetCurMinBlockY() {
		return curMinBlockY_;
	}

	/**
	 * Process the line-clear sequences
	 * @param msec Amount of time elapsed from the last execution (in milliseconds)
	 */
	public void Run(int msec) {
		curTime_ += msec;

		if(IsLineAnimating()) {
			animatingLine_.lifeLimitMSec_ -= msec;

			if(animatingLine_.lifeLimitMSec_ <= 0) {
				animatingLine_.lifeLimitMSec_ = 0;
				for(int i = 0; i < animatingLine_.vecIdx_.size(); i++) {
					fallingLine_.vecIdx_.addElement(animatingLine_.vecIdx_.get(i));
				}
				//fallingLine_.vecIdx_.addAll(animatingLine_.vecIdx_);
				fallingLine_.ratio_ = 0f;
				for(int i = fallingLine_.vecIdx_.size() - 1; i >= 0; i--) {
					ClearLine(fallingLine_.vecIdx_.get(i));
				}
				animatingLine_.vecIdx_.clear();
			}
		}

		if(IsLineFalling()) {
			if(gamePlay_.GetPlayerSetting().MinoGravity != 0) {
				float spd = (1f / gamePlay_.GetPlayerSetting().MinoGravity) * msec;
				fallingLine_.ratio_ += spd;
			} else {
				fallingLine_.ratio_ = 100f;
			}

			while(true) {
				if( !(fallingLine_.vecIdx_.size() > 0 && fallingLine_.ratio_ >= 1f) ) {
					break;
				}
				fallingLine_.ratio_ -= 1f;
				RemoveLine(fallingLine_.vecIdx_.get(0));
				for(int i = 0; i < fallingLine_.vecIdx_.size(); i++) {
					int temp = fallingLine_.vecIdx_.get(i);
					fallingLine_.vecIdx_.set(i, temp + 1);
				}
				fallingLine_.vecIdx_.remove(0);
			}
		}
	}

	/**
	 * Grid Data<br>
	 * [Anonymous class]
	 */
	public static class Grid {
		/** Mino */
		public Mino mino_ = new Mino();

		/** <code>true</code> if empty */
		public boolean empty_ = true;

		/** Creation time of this grid */
		public int createTime_ = 0;
	}

	/**
	 * Detailed results of PutTetrimino<br>
	 * [Anonymous class]
	 */
	public static class PutTetriminoInfos {
		/** Number of rows the Tetrimino passed (Only used when the drop parameter is used) */
		public int downLineCnt = 0;

		/** <code>true</code> if T-Spin */
		public boolean isTSpin = false;

		/** <code>true</code> if mini-spin */
		public boolean isCheapSpin = false;

		/** <code>true</code> if the Tetrimino has locked to the non-visible portion of the playfield
		    (It will cause lockout death) */
		public boolean lockOut = false;
	}

	/**
	 * Animation status for cleared-lines<br>
	 * [Anonymous class]
	 */
	public static class AnimatingLine {
		/** Amount of time remaining before the lines disappear */
		public int lifeLimitMSec_ = 0;

		/** Dynamic array of Y positions */
		public Vector<Integer> vecIdx_ = new Vector<Integer>();
	}

	/**
	 * Falling stack status<br>
	 * [Anonymous class]
	 */
	public static class FallingLine {
		/** Dynamic array of Y positions */
		public Vector<Integer> vecIdx_ = new Vector<Integer>();

		/** Fall speed */
		public float ratio_ = 0f;
	}
}
