package tk.to_p.blankedoutbox.garlicclone.gui.awt;


import java.awt.Color;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Random;

import tk.to_p.blankedoutbox.garlicclone.ai.LuaTetrisAI;
import tk.to_p.blankedoutbox.garlicclone.ai.TetrisAI;
import tk.to_p.blankedoutbox.garlicclone.core.EventReceiver;
import tk.to_p.blankedoutbox.garlicclone.core.GamePlay;
import tk.to_p.blankedoutbox.garlicclone.core.GarlicConst;
import tk.to_p.blankedoutbox.garlicclone.core.Mino;
import tk.to_p.blankedoutbox.garlicclone.core.Player;
import tk.to_p.blankedoutbox.garlicclone.core.Playfield;
import tk.to_p.blankedoutbox.garlicclone.core.TetriMinos;
import cx.it.nullpo.nm8.util.CustomProperties;

/**
 * AWT frontend of GarlicClone.
 * Please note I didn't port rendering routines to Java; I created from scratch.
 */
public class GarlicCloneAWT extends Frame implements Runnable {
	/** Serial version UID */
	private static final long serialVersionUID = 2L;

	/** Version string */
	public static final String VERSION_STRING = "1.1";

	/** Game type constants: Marathon */
	public static final int GAMETYPE_MARATHON = 0;
	/** Game type constants: Sprint */
	public static final int GAMETYPE_SPRINT = 1;
	/** Game type constants: Ultra */
	public static final int GAMETYPE_ULTRA = 2;

	/** Config (Thanks NullpoMino Team for CustomProperties) */
	public CustomProperties propConfig;

	/** Main game engine */
	public GamePlay gamePlay;

	/** AI for Auto Play */
	public TetrisAI ai;

	/** AI for Anti Miss */
	public TetrisAI antimissAI;

	protected boolean antimissReady;
	protected boolean antimissMissed;
	protected int antimissCheckedId;
	protected int antimissEffectFrame;

	/** Size of window "border" */
	public Insets insets;

	/** Thread */
	public Thread thread;

	/** Button input flags */
	public int buttonFlag;

	/** Image for double-buffering */
	protected Image dbImage;

	/** Graphics for double-buffering */
	protected Graphics dbGraphics;

	/** Blocks image */
	protected Image imgBlock;

	/** Frame image */
	protected Image imgFrame;

	/** Game has ended flag */
	protected boolean gameEnded;

	/** Thread sleep time */
	protected long sleepTime;

	/** Start level */
	protected int startLevel;

	/** Game type */
	protected int gameType;

	/** Marathon: Required lines to level up */
	protected int marathonLevelUpLines;

	/** Marathon: Required lines to win */
	protected int marathonGoal;

	/** Sprint: Required lines to win */
	protected int sprintGoal;

	/** Ultra: Required time to win */
	protected int ultraTime;

	/** Marathon: Level up flag */
	protected boolean marathonLevelUpFlag;

	/**
	 * Main
	 * @param args Command-line params
	 */
	public static void main(String[] args) {
		new GarlicCloneAWT();
	}

	/**
	 * Constructor
	 */
	public GarlicCloneAWT() {
		setTitle("GarlicClone v" + VERSION_STRING);
		setVisible(true);

		insets = getInsets();
		setSize(376 + insets.left + insets.right, 450 + insets.top + insets.bottom);
		setResizable(false);

		propConfig = CustomProperties.load("config/config.properties");
		if(propConfig == null) propConfig = new CustomProperties();

		imgBlock = Toolkit.getDefaultToolkit().getImage("res/image/block.png");
		imgFrame = Toolkit.getDefaultToolkit().getImage("res/image/frame.png");

		GarlicConst.InitGarlic();
		gamePlay = new GamePlay();
		gamePlay.eventReceivers_.add(new EventReceiverTest());

		// Game config
		sleepTime = propConfig.getProperty("game.sleepTime", 10L);
		startLevel = propConfig.getProperty("game.level", 1);
		gamePlay.GetStatistics().curLevel_ = startLevel;
		gameType = propConfig.getProperty("game.type", GAMETYPE_MARATHON);
		marathonLevelUpLines = propConfig.getProperty("game.marathon.levelUpLines", 10);
		marathonGoal = propConfig.getProperty("game.marathon.goal", 200);
		sprintGoal = propConfig.getProperty("game.sprint.goal", 40);
		ultraTime = propConfig.getProperty("game.ultra.time", 2);

		// Speed settings
		gamePlay.GetPlayerSetting().MoveDelay = propConfig.getProperty("speed.moveDelay", 130);
		gamePlay.GetPlayerSetting().MoveInterval = propConfig.getProperty("speed.moveInterval", 15);
		gamePlay.GetPlayerSetting().SoftDropDelay = propConfig.getProperty("speed.softDropDelay", 10);
		gamePlay.GetPlayerSetting().SoftDropInterval = propConfig.getProperty("speed.softDropInterval", 10);
		gamePlay.GetPlayerSetting().ExtendedPlacementCount = propConfig.getProperty("speed.extendedPlacementCount", 15);
		gamePlay.GetPlayerSetting().LockDownDelay = propConfig.getProperty("speed.lockDownDelay", 500);
		gamePlay.GetPlayerSetting().MinoGravity = propConfig.getProperty("speed.minoGravity", 50);
		gamePlay.GetPlayerSetting().LineLifeTime = propConfig.getProperty("speed.lineLifeTime", 0);

		// Setup an AI
		int aiType = propConfig.getProperty("ai.type", 0);
		if(aiType == 1) {
			ai = new TetrisAI(gamePlay);
		} else if(aiType == 2) {
			String aiFileName = propConfig.getProperty("ai.lua.file", "res/ai/test_tspin.lua");
			System.out.println("Loading AI from " + aiFileName);
			ai = new LuaTetrisAI(gamePlay, aiFileName);
		} else {
			ai = null;
		}

		// Think/Move delay for AI
		if(ai != null) {
			ai.reqThinkDelayFrame_ = propConfig.getProperty("ai.thinkDelay", 0);
			ai.reqMoveDelayFrame_ = propConfig.getProperty("ai.moveDelay", 0);
		}

		// Anti-miss AI
		antimissEffectFrame = 0;
		boolean enableAntiMiss = propConfig.getProperty("antimiss.enable", false);
		if(enableAntiMiss) {
			String aiFileName = propConfig.getProperty("antimiss.lua.file", "res/ai/test_dropchecker.lua");
			System.out.println("Loading Anti-miss AI from " + aiFileName);
			antimissAI = new LuaTetrisAI(gamePlay, aiFileName);
		} else {
			antimissAI = null;
		}

		buttonFlag = 0;
		gameEnded = false;
		marathonLevelUpFlag = false;

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosed(WindowEvent e) {
				System.exit(0);
			}

			@Override
			public void windowClosing(WindowEvent e) {
				dispose();
			}
		});

		addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				int k = e.getKeyCode();

				if(gamePlay.IsStartGame() && !gamePlay.IsGameOver() && !gamePlay.IsPauseGame()) {
					switch(k) {
					case KeyEvent.VK_LEFT:
					case KeyEvent.VK_NUMPAD4:
						buttonFlag |= GarlicConst.ControllerFlag_LEFT;
						break;
					case KeyEvent.VK_RIGHT:
					case KeyEvent.VK_NUMPAD6:
						buttonFlag |= GarlicConst.ControllerFlag_RIGHT;
						break;
					case KeyEvent.VK_SPACE:
					case KeyEvent.VK_NUMPAD8:
						buttonFlag |= GarlicConst.ControllerFlag_HARD;
						break;
					case KeyEvent.VK_DOWN:
					case KeyEvent.VK_NUMPAD2:
						buttonFlag |= GarlicConst.ControllerFlag_SOFT;
						break;
					case KeyEvent.VK_UP:
					case KeyEvent.VK_X:
					case KeyEvent.VK_NUMPAD1:
					case KeyEvent.VK_NUMPAD5:
					case KeyEvent.VK_NUMPAD9:
						buttonFlag |= GarlicConst.ControllerFlag_CW;
						break;
					case KeyEvent.VK_Z:
					case KeyEvent.VK_CONTROL:
					case KeyEvent.VK_NUMPAD3:
					case KeyEvent.VK_NUMPAD7:
						buttonFlag |= GarlicConst.ControllerFlag_CCW;
						break;
					case KeyEvent.VK_C:
					case KeyEvent.VK_SHIFT:
					case KeyEvent.VK_NUMPAD0:
						buttonFlag |= GarlicConst.ControllerFlag_HOLD;
						break;
					case KeyEvent.VK_ESCAPE:
						gamePlay.PauseGame(true);
						break;
					}
				} else {
					if((k == KeyEvent.VK_ESCAPE) && gamePlay.IsPauseGame()) {
						gamePlay.PauseGame(false);
					}
					if((k == KeyEvent.VK_LEFT) && !gamePlay.IsPauseGame()) {
						gameType--;
						if(gameType < 0) gameType = 2;
					}
					if((k == KeyEvent.VK_RIGHT) && !gamePlay.IsPauseGame()) {
						gameType++;
						if(gameType > 2) gameType = 0;
					}
					if(k == KeyEvent.VK_ENTER) {
						buttonFlag = 0;
						gameEnded = false;
						marathonLevelUpFlag = false;

						Random rand = new Random();
						gamePlay.Init();
						gamePlay.GetPlayer().bag_.tetriMinosInBag_.clear();
						gamePlay.StartGame(rand.nextLong(), 1000);
					}
				}
			}

			@Override
			public void keyReleased(KeyEvent e) {
				int k = e.getKeyCode();

				if(gamePlay.IsStartGame() && !gamePlay.IsGameOver()) {
					switch(k) {
					case KeyEvent.VK_LEFT:
					case KeyEvent.VK_NUMPAD4:
						buttonFlag &= ~GarlicConst.ControllerFlag_LEFT;
						break;
					case KeyEvent.VK_RIGHT:
					case KeyEvent.VK_NUMPAD6:
						buttonFlag &= ~GarlicConst.ControllerFlag_RIGHT;
						break;
					case KeyEvent.VK_SPACE:
					case KeyEvent.VK_NUMPAD8:
						buttonFlag &= ~GarlicConst.ControllerFlag_HARD;
						break;
					case KeyEvent.VK_DOWN:
					case KeyEvent.VK_NUMPAD2:
						buttonFlag &= ~GarlicConst.ControllerFlag_SOFT;
						break;
					case KeyEvent.VK_UP:
					case KeyEvent.VK_X:
					case KeyEvent.VK_NUMPAD1:
					case KeyEvent.VK_NUMPAD5:
					case KeyEvent.VK_NUMPAD9:
						buttonFlag &= ~GarlicConst.ControllerFlag_CW;
						break;
					case KeyEvent.VK_Z:
					case KeyEvent.VK_CONTROL:
					case KeyEvent.VK_NUMPAD3:
					case KeyEvent.VK_NUMPAD7:
						buttonFlag &= ~GarlicConst.ControllerFlag_CCW;
						break;
					case KeyEvent.VK_C:
					case KeyEvent.VK_SHIFT:
					case KeyEvent.VK_NUMPAD0:
						buttonFlag &= ~GarlicConst.ControllerFlag_HOLD;
						break;
					}
				}
			}
		});

		thread = new Thread(this);
		thread.setDaemon(true);
		thread.start();
	}

	/*
	 * Game Thread
	 */
	public void run() {
		long lastExecTime = System.nanoTime();

		while(true) {
			synchronized (gamePlay) {
				if(gamePlay.IsStartGame() && !gamePlay.IsGameOver()) {
					gamePlay.SetControllerFlags(buttonFlag);

					long currentTime = System.nanoTime();
					int loopTime = (int)((currentTime - lastExecTime) / 1000000);

					if(isFocused()) {
						if(ai != null) {
							try {
								ai.Run();
								buttonFlag = gamePlay.controller_.controllerFlags_;
							} catch (Exception e) {
								System.err.println("AI bug");
								e.printStackTrace();
							}
						}

						if(antimissAI != null) {
							try {
								if((buttonFlag & GarlicConst.ControllerFlag_HARD) != 0) {
									Player.Piece piece = gamePlay.GetPlayer().GetCurPiece();

									if(antimissReady && (piece != null)) {
										antimissReady = false;
										if(antimissCheckedId != piece.id_) {
											if(!antimissMissed && antimissAI.IsMissDrop(1, 100, 0)) {
												antimissMissed = true;
												antimissEffectFrame = 1000;
											}
											antimissCheckedId = piece.id_;
										}
									}

									if(antimissMissed) {
										buttonFlag &= ~GarlicConst.ControllerFlag_HARD;
									}
								} else {
									antimissReady = true;
									antimissMissed = false;
								}

								gamePlay.SetControllerFlags(buttonFlag);
							} catch (Exception e) {
								System.err.println("Antimiss AI bug");
								e.printStackTrace();
							}

							if(antimissEffectFrame > 0) antimissEffectFrame -= loopTime;
						}

						try {
							gamePlay.Run(loopTime);
							render();
						} catch (NullPointerException e) {}
					}

					lastExecTime = currentTime;
				} else {
					if(isFocused()) {
						try {
							render();
						} catch (NullPointerException e) {}
					}

					lastExecTime = System.nanoTime();
				}
			}

			if(sleepTime > 0) {
				try {
					Thread.sleep(sleepTime);
				} catch (InterruptedException e) {}
			}
		}
	}

	/**
	 * Draw the screen
	 */
	public void render() {
		if(dbImage == null) {
			dbImage = createImage(376, 450);
			dbGraphics = dbImage.getGraphics();
		}

		dbGraphics.setColor(Color.LIGHT_GRAY);
		dbGraphics.fillRect(0, 0, 376, 450);

		dbGraphics.drawImage(imgFrame, 0, 0, null);

		// Playfield
		if(gamePlay.GetPlayfield() != null) {
			for(int x = 0; x < gamePlay.GetPlayfield().GetWidth(); x++) {
				for(int y = 0; y < gamePlay.GetPlayfield().GetHeight() / 2; y++) {
					Playfield.Grid grid = gamePlay.GetPlayfield().GetGrid(x, y + (gamePlay.GetPlayfield().GetHeight() / 2));
					float addY = gamePlay.GetPlayfield().GetLineAddY(y + (gamePlay.GetPlayfield().GetHeight() / 2)) * 18f;

					if(grid != null && !grid.empty_) {
						drawMino(dbGraphics, 93 + x * 18, 43 + y * 18 + (int)addY, 2, grid.mino_);
					}
				}
			}
		}

		// Piece preview
		if(gamePlay.IsStartGame() || gameEnded) {
			for(int i = 0; i < Math.min(gamePlay.GetPlayerSetting().NextBlockCount, gamePlay.GetPlayerSetting().BagSize); i++) {
				int id = gamePlay.GetPlayer().GetTetriminoInBag(i);
				drawTetriMino(dbGraphics, 286, 81 + i * 54, false, GarlicConst.TetriMinos_[id].tetriMinos_[0]);
			}
		}

		// Hold piece
		if(gamePlay.GetPlayer().GetTetriminoInHold() != GarlicConst.TetriminoType_None) {
			drawTetriMino(dbGraphics, 8, 81, gamePlay.GetPlayer().CanHold() ? 0 : 2, GarlicConst.TetriMinos_[gamePlay.GetPlayer().GetTetriminoInHold()].tetriMinos_[0]);
		}

		// Ready?
		if(gamePlay.IsReadyTime()) {
			dbGraphics.setColor(Color.yellow);
			dbGraphics.drawString("READY?", 10, 310);
		}

		// Score
		GamePlay.Statistics stats = gamePlay.GetStatistics();
		if(stats != null) {
			dbGraphics.setColor(Color.orange);
			dbGraphics.drawString("SCORE", 10, 160);
			dbGraphics.setColor(Color.white);
			dbGraphics.drawString("" + stats.curScore_, 10, 175);
			int getPoints = stats.thisFilledScore_ + stats.thisComboScore_ + stats.thisTSpinScore_ + stats.thisPerfectClearScore_;
			if(getPoints > 0) dbGraphics.drawString("(+" + getPoints + ")", 10, 190);

			dbGraphics.setColor(Color.orange);
			dbGraphics.drawString("LINE", 10, 205);
			dbGraphics.setColor(Color.white);
			int lines = stats.totalFilledLines_;
			if(gameType == GAMETYPE_SPRINT) {
				lines = sprintGoal - stats.totalFilledLines_;
				if(lines < 0) lines = 0;

				if(lines < 30) dbGraphics.setColor(Color.yellow);
				if(lines < 20) dbGraphics.setColor(Color.orange);
				if(lines < 10) dbGraphics.setColor(Color.red);
			}
			dbGraphics.drawString("" + lines, 10, 220);

			dbGraphics.setColor(Color.orange);
			dbGraphics.drawString("LEVEL", 10, 235);
			dbGraphics.setColor(Color.white);
			dbGraphics.drawString("" + stats.curLevel_, 10, 250);

			dbGraphics.setColor(Color.orange);
			dbGraphics.drawString("TIME", 10, 265);
			dbGraphics.setColor(Color.white);
			int time = gamePlay.GetPlayTime();
			if(gameType == GAMETYPE_ULTRA) {
				time = (ultraTime * 60 * 1000) - gamePlay.GetPlayTime();
				if(time < 0) time = 0;

				if(time < 60 * 1000) dbGraphics.setColor(Color.yellow);
				if(time < 30 * 1000) dbGraphics.setColor(Color.orange);
				if(time < 15 * 1000) dbGraphics.setColor(Color.red);
			}
			dbGraphics.drawString(getTimeAsString(time), 10, 280);

			dbGraphics.setColor(Color.yellow);
			if(stats.IsB2B()) {
				dbGraphics.drawString("Back-to-Back", 10, 310);
			}
			if(stats.IsTSpin()) {
				if(stats.lastFilledLines_ == 0) {
					dbGraphics.drawString("T-Spin", 10, 325);
				} else if(stats.lastFilledLines_ == 1) {
					if(stats.lastCheapSpin_) {
						dbGraphics.drawString("T-Spin Mini", 10, 325);
					} else {
						dbGraphics.drawString("T-Spin Single", 10, 325);
					}
				} else if(stats.lastFilledLines_ == 2) {
					dbGraphics.drawString("T-Spin Double", 10, 325);
				} else if(stats.lastFilledLines_ == 3) {
					dbGraphics.drawString("T-Spin Triple", 10, 325);
				}
			} else if(stats.lastFilledLines_ == 1) {
				dbGraphics.drawString("Single", 10, 325);
			} else if(stats.lastFilledLines_ == 2) {
				dbGraphics.drawString("Double", 10, 325);
			} else if(stats.lastFilledLines_ == 3) {
				dbGraphics.drawString("Triple", 10, 325);
			} else if(stats.lastFilledLines_ == 4) {
				dbGraphics.drawString("Te*ris", 10, 325);
			}
			if(stats.curComboCnt_ >= 2) {
				dbGraphics.drawString((stats.curComboCnt_ - 1) + " REN", 10, 340);
			}
			if(stats.perfectClear_) {
				dbGraphics.drawString("PERFECT", 10, 355);
				dbGraphics.drawString("CLEAR!!", 20, 370);
			} else if(marathonLevelUpFlag) {
				dbGraphics.drawString("LEVEL UP!", 10, 355);
			}

			if(antimissEffectFrame > 0) {
				dbGraphics.setColor(Color.pink);
				dbGraphics.drawString("ANTI-MISS", 10, 380);
			}

			if( (gamePlay.IsStartGame() && !gamePlay.IsGameOver()) || gamePlay.IsPauseGame() ) {
				dbGraphics.setColor(Color.white);
			} else {
				dbGraphics.setColor(Color.green);
			}
			if(gameType == GAMETYPE_MARATHON) dbGraphics.drawString("MARATHON", 10, 400);
			if(gameType == GAMETYPE_SPRINT) dbGraphics.drawString("SPRINT", 10, 400);
			if(gameType == GAMETYPE_ULTRA) dbGraphics.drawString("ULTRA", 10, 400);
		}

		// Game Paused
		if(gamePlay.IsPauseGame()) {
			dbGraphics.setColor(Color.green);
			dbGraphics.drawString("Press ESC key to resume", 10, 20);
			dbGraphics.drawString("Press ENTER key to restart", 10, 35);
		}

		if(gamePlay.IsStartGame() && !gamePlay.IsGameOver()) {
			// Current piece
			if(gamePlay.GetPlayer().GetCurPiece() != null) {
				Player.Piece piece = gamePlay.GetPlayer().GetCurPiece();

				if(gamePlay.GetPlayerSetting().DrawGhost) {
					drawCurrentTetriMino(dbGraphics, 93 + piece.xGhost_ * 18, 43 + (piece.yGhost_ - gamePlay.GetPlayfield().GetHeight() / 2) * 18,
							true, piece);
				}

				drawCurrentTetriMino(dbGraphics, 93 + piece.x_ * 18, 43 + (piece.y_ - gamePlay.GetPlayfield().GetHeight() / 2) * 18,
						false, piece);
			}
		} else {
			if(gameEnded) {
				// Game Over
				dbGraphics.setColor(Color.green);
				dbGraphics.drawString("Press ENTER key to restart", 10, 20);

				GamePlay.Statistics stats2 = gamePlay.GetStatistics();
				if(stats2 != null) {
					dbGraphics.setColor(Color.cyan);
					String strStats = "";
					strStats += "Single:" + stats2.filledCount_[0];
					strStats += ", Double:" + stats2.filledCount_[1];
					strStats += ", Triple:" + stats2.filledCount_[2];
					strStats += ", Te*ris:" + stats2.filledCount_[3];
					strStats += ", Max REN:" + ((stats2.maxComboCnt_ < 0) ? "N/A" : ""+(stats2.maxComboCnt_ - 1));
					strStats += ", T-Spin:" + stats2.tspinCnt_;
					dbGraphics.drawString(strStats, 10, 35);
				}
			} else {
				// Startup
				dbGraphics.setColor(Color.green);
				dbGraphics.drawString("Press ENTER key to play", 10, 20);
				dbGraphics.drawString("Press Left or Right key to select mode", 10, 35);
			}
		}

		Graphics g = getGraphics();
		if(g != null) {
			g.drawImage(dbImage, insets.left, insets.top, null);
			g.dispose();
		}
	}

	/**
	 * Draw a Mino
	 * @param g Graphics to draw
	 * @param x X position
	 * @param y Y position
	 * @param ghost <code>true</code> to draw ghost mino
	 * @param mino Mino object
	 */
	protected void drawMino(Graphics g, int x, int y, boolean ghost, Mino mino) {
		drawMino(g, x, y, ghost ? 1 : 0, mino);
	}

	protected void drawMino(Graphics g, int x, int y, int type, Mino mino) {
		int x2 = x + (mino.x_ * 18);
		int y2 = y + (mino.y_ * 18);
		int color = mino.color_ - 1;

		if(type == 2) {
			g.drawImage(imgBlock, x2, y2, x2 + 18, y2 + 18, color * 18, 36, color * 18 + 18, 54, null);
		} else if(type == 1) {
			g.drawImage(imgBlock, x2, y2, x2 + 18, y2 + 18, color * 18, 18, color * 18 + 18, 36, null);
		} else {
			g.drawImage(imgBlock, x2, y2, x2 + 18, y2 + 18, color * 18, 0, color * 18 + 18, 18, null);
		}
	}

	/**
	 * Draw a Tetrimino
	 * @param g Graphics to draw
	 * @param x X position
	 * @param y Y position
	 * @param ghost <code>true</code> to draw ghost mino
	 * @param tetriMino TetriMino object
	 */
	protected void drawTetriMino(Graphics g, int x, int y, boolean ghost, TetriMinos.TetriMino tetriMino) {
		for(int i = 0; i < tetriMino.minos_.length; i++) {
			if(ghost)
				drawMino(g, x, y, 1, tetriMino.minos_[i]);
			else
				drawMino(g, x, y, 0, tetriMino.minos_[i]);
		}
	}

	protected void drawTetriMino(Graphics g, int x, int y, int type, TetriMinos.TetriMino tetriMino) {
		for(int i = 0; i < tetriMino.minos_.length; i++) {
			drawMino(g, x, y, type, tetriMino.minos_[i]);
		}
	}

	/**
	 * Draw current Tetrimino
	 * @param g Graphics to draw
	 * @param x X position
	 * @param y Y position
	 * @param ghost <code>true</code> to draw ghost mino
	 * @param piece Piece object
	 */
	protected void drawCurrentTetriMino(Graphics g, int x, int y, boolean ghost, Player.Piece piece) {
		TetriMinos.TetriMino tetriMino = piece.GetTetriMino();
		int pieceY = ghost ? piece.yGhost_ : piece.y_;

		for(int i = 0; i < tetriMino.minos_.length; i++) {
			if(pieceY + tetriMino.minos_[i].y_ >= 20) {
				drawMino(g, x, y, ghost ? 1 : 0, tetriMino.minos_[i]);
			}
		}
	}

	/**
	 * Convert play time to String
	 * @param t Play Time
	 * @return Play Time in String format (00:11.222)
	 */
	protected String getTimeAsString(int t) {
		int minute = 0;
		int second = 0;
		int millisecond = 0;

		while(t >= 1000 * 60) {
			t -= 1000 * 60;
			minute++;
		}
		while(t >= 1000) {
			t -= 1000;
			second++;
		}
		millisecond = t;

		return String.format("%02d:%02d.%03d", minute, second, millisecond);
	}

	/**
	 * EventReceiver for level-up
	 */
	private class EventReceiverTest extends EventReceiver {
		@Override
		public void EndGame() {
			gameEnded = true;

			if(ai != null) {
				ai.Shutdown();
			}
		}

		@Override
		public void Ready(int delay) {
			gamePlay.GetStatistics().curLevel_ = startLevel;
		}

		@Override
		public void GameOver(int reason) {
			gamePlay.EndGame();
		}

		@Override
		public void EndPiece(Player.Piece piece) {
			marathonLevelUpFlag = false;

			// Marathon
			if(gameType == GAMETYPE_MARATHON) {
				if( (marathonLevelUpLines > 0) &&
					(gamePlay.GetStatistics().totalFilledLines_ >= gamePlay.GetStatistics().curLevel_ * marathonLevelUpLines) &&
					(gamePlay.GetStatistics().curLevel_ < 20) )
				{
					gamePlay.GetStatistics().curLevel_++;
					marathonLevelUpFlag = true;
				}

				if((marathonGoal > 0) && (gamePlay.GetStatistics().totalFilledLines_ >= marathonGoal)) {
					gamePlay.EndGame();
				}
			}

			// Sprint
			if(gameType == GAMETYPE_SPRINT) {
				if((sprintGoal > 0) && (gamePlay.GetStatistics().totalFilledLines_ >= sprintGoal)) {
					gamePlay.EndGame();
				}
			}
		}

		@Override
		public void OnUpdatePlayTime(int time) {
			// Ultra
			if(gameType == GAMETYPE_ULTRA) {
				if(time >= ultraTime * 60 * 1000) {
					gamePlay.EndGame();
				}
			}
		}
	}
}
